'''
Functions for looking up element names, numbers, and symbols
and for converting between angular momentum conventions

This module has functions for looking up element information by
symbol, name, or number. It also has functions for converting
angular momentum between integers (0, 1, 2) and letters (s, p, d).
'''

# Contains the symbols, names, and Z numbers for all known elements
# NOTE: Some Z number have multiple entries. This is to allow for querying
#       based on older, systematic names/symbols (like uuh), however a query
#       of the Z number should return the 'new', official name (livermorium or lv)
#       This also allows for different spellings of some names (cesium/caesium, aluminum/aluminium)
# yapf: disable
_data_table = [
    ("h",     1, "hydrogen"),      ("he",    2, "helium"),      ("li",    3, "lithium"),
    ("be",    4, "beryllium"),     ("b",     5, "boron"),       ("c",     6, "carbon"),
    ("n",     7, "nitrogen"),      ("o",     8, "oxygen"),      ("f",     9, "fluorine"),
    ("ne",   10, "neon"),          ("na",   11, "sodium"),      ("mg",   12, "magnesium"),
    ("al",   13, "aluminum"),      ("al",   13, "aluminium"),   ("si",   14, "silicon"),
    ("p",    15, "phosphorus"),    ("s",    16, "sulfur"),      ("cl",   17, "chlorine"),
    ("ar",   18, "argon"),         ("k",    19, "potassium"),   ("ca",   20, "calcium"),
    ("sc",   21, "scandium"),      ("ti",   22, "titanium"),    ("v",    23, "vanadium"),
    ("cr",   24, "chromium"),      ("mn",   25, "manganese"),   ("fe",   26, "iron"),
    ("co",   27, "cobalt"),        ("ni",   28, "nickel"),      ("cu",   29, "copper"),
    ("zn",   30, "zinc"),          ("ga",   31, "gallium"),     ("ge",   32, "germanium"),
    ("as",   33, "arsenic"),       ("se",   34, "selenium"),    ("br",   35, "bromine"),
    ("kr",   36, "krypton"),       ("rb",   37, "rubidium"),    ("sr",   38, "strontium"),
    ("y",    39, "yttrium"),       ("zr",   40, "zirconium"),   ("nb",   41, "niobium"),
    ("mo",   42, "molybdenum"),    ("tc",   43, "technetium"),  ("ru",   44, "ruthenium"),
    ("rh",   45, "rhodium"),       ("pd",   46, "palladium"),   ("ag",   47, "silver"),
    ("cd",   48, "cadmium"),       ("in",   49, "indium"),      ("sn",   50, "tin"),
    ("sb",   51, "antimony"),      ("te",   52, "tellurium"),   ("i",    53, "iodine"),
    ("xe",   54, "xenon"),         ("cs",   55, "cesium"),      ("cs",   55, "caesium"),
    ("ba",   56, "barium"),        ("la",   57, "lanthanum"),   ("ce",   58, "cerium"),
    ("pr",   59, "praseodymium"),  ("nd",   60, "neodymium"),   ("pm",   61, "promethium"),
    ("sm",   62, "samarium"),      ("eu",   63, "europium"),    ("gd",   64, "gadolinium"),
    ("tb",   65, "terbium"),       ("dy",   66, "dysprosium"),  ("ho",   67, "holmium"),
    ("er",   68, "erbium"),        ("tm",   69, "thulium"),     ("yb",   70, "ytterbium"),
    ("lu",   71, "lutetium"),      ("hf",   72, "hafnium"),     ("ta",   73, "tantalum"),
    ("w",    74, "tungsten"),      ("re",   75, "rhenium"),     ("os",   76, "osmium"),
    ("ir",   77, "iridium"),       ("pt",   78, "platinum"),    ("au",   79, "gold"),
    ("hg",   80, "mercury"),       ("tl",   81, "thallium"),    ("pb",   82, "lead"),
    ("bi",   83, "bismuth"),       ("po",   84, "polonium"),    ("at",   85, "astatine"),
    ("rn",   86, "radon"),         ("fr",   87, "francium"),    ("ra",   88, "radium"),
    ("ac",   89, "actinium"),      ("th",   90, "thorium"),     ("pa",   91, "protactinium"),
    ("u",    92, "uranium"),       ("np",   93, "neptunium"),   ("pu",   94, "plutonium"),
    ("am",   95, "americium"),     ("cm",   96, "curium"),      ("bk",   97, "berkelium"),
    ("cf",   98, "californium"),   ("es",   99, "einsteinium"), ("fm",  100, "fermium"),
    ("md",  101, "mendelevium"),   ("no",  102, "nobelium"),    ("lr",  103, "lawrencium"),
    ("rf",  104, "rutherfordium"), ("ha",  105, "hahnium"),     ("db",  105, "dubnium"),
    ("sg",  106, "seaborgium"),    ("bh",  107, "bohrium"),     ("hs",  108, "hassium"),
    ("mt",  109, "meitnerium"),    ("uun", 110, "ununnilium"),  ("ds",  110, "darmstadtium"),
    ("uuu", 111, "unununium"),     ("rg",  111, "roentgenium"), ("uub", 112, "ununbium"),
    ("cn",  112, "copernicium"),   ("uut", 113, "ununtrium"),   ("nh",  113, "nihonium"),
    ("uuq", 114, "ununquadium"),   ("fl",  114, "flerovium"),   ("uup", 115, "ununpentium"),
    ("mc",  115, "moscovium"),     ("uuh", 116, "ununhexium"),  ("lv",  116, "livermorium"),
    ("uus", 117, "ununseptium"),   ("ts",  117, "tennessine"),  ("uuo", 118, "ununoctium"),
    ("og",  118, "oganesson"),     ("uue", 119, "ununennium"),  ("ubn", 120, "unbinilium")
]
# yapf: enable

# Maps Z to element data
# Note that only the last entry for a particular Z number will
# be stored (which contains the new, official data rather than
# the older systematic names for some elements)
_element_Z_map = {x[1]: x for x in _data_table}

# Maps element symbol to element data
_element_sym_map = {x[0]: x for x in _data_table}

# Maps element name to element data
_element_name_map = {x[2]: x for x in _data_table}

# Maps AM characters to integers (the integer is the
# index of the character in this string)
# This is the 'hik' convention, where AM=7 is k
_amchar_map_hik = 'spdfghiklmnoqrtuvwxyzabce'

# This is the 'hij' convention, where AM=7 is j
_amchar_map_hij = 'spdfghijklmnoqrtuvwxyzabce'


def all_element_names():
    '''Obtain a list of the names of all the elements'''
    return list(_element_name_map.keys())


def element_data_from_Z(Z):
    '''Obtain elemental data given a Z number

    An exception is thrown if the Z number is not found
    '''

    # Z may be a str
    if isinstance(Z, str) and Z.isdecimal():
        Z = int(Z)

    if Z not in _element_Z_map:
        raise KeyError('No element data for Z = {}'.format(Z))
    return _element_Z_map[Z]


def element_data_from_sym(sym):
    '''Obtain elemental data given an elemental symbol

    The given symbol is not case sensitive

    An exception is thrown if the symbol is not found
    '''

    sym_lower = sym.lower()
    if sym_lower not in _element_sym_map:
        raise KeyError('No element data for symbol \'{}\''.format(sym))
    return _element_sym_map[sym_lower]


def element_data_from_name(name):
    '''Obtain elemental data given an elemental name

    The given name is not case sensitive

    An exception is thrown if the name is not found
    '''

    name_lower = name.lower()
    if name_lower not in _element_name_map:
        raise KeyError('No element data for name \'{}\''.format(name))
    return _element_name_map[name_lower]


def element_name_from_Z(Z, normalize=False):
    '''Obtain an element's name from its Z number

    An exception is thrown if the Z number is not found

    If normalize is True, the first letter will be capitalized
    '''

    r = element_data_from_Z(Z)[2]
    if normalize:
        return r.capitalize()
    else:
        return r


def element_Z_from_name(name, as_str=False):
    '''Obtain an element's Z number given its name

    If as_str is True, then a string is returned (ie, '1' for Hydrogen)

    An exception is thrown if the name is not found
    '''

    Z = element_data_from_name(name)[1]
    if as_str:
        Z = str(Z)
    return Z


def element_sym_from_Z(Z, normalize=False):
    '''Obtain an element's symbol from its Z number

    An exception is thrown if the Z number is not found

    If normalize is True, the first letter will be capitalized
    '''

    r = element_data_from_Z(Z)[0]
    if normalize:
        return r.capitalize()
    else:
        return r


def element_Z_from_sym(sym, as_str=False):
    '''Obtain an element's Z-number given its symbol

    If as_str is True, then a string is returned (ie, '1' for Hydrogen)

    An exception is thrown if the symbol is not found
    '''

    Z = element_data_from_sym(sym)[1]
    if as_str:
        Z = str(Z)
    return Z


def amint_to_char(am, hij=False, use_L=False):
    '''Convert an angular momentum integer to a character

    The input is a list (to handle sp, spd, ... orbitals). The return
    value is a string

    For example, converts [0] to 's' and [0,1,2] to 'spd'

    If hij is True, the ordering spdfghijkl is used. Otherwise, the
    ordering will be spdfghikl (skipping j)

    If use_L is True, sp shells ([0,1]) will return l instead
    '''

    if use_L and am == [0, 1]:
        return 'l'
    if hij:
        amchar_map = _amchar_map_hij
    else:
        amchar_map = _amchar_map_hik

    amchar = []

    for a in am:
        if a < 0:
            raise IndexError('Angular momentum must be a positive integer (not {})'.format(a))
        if a >= len(amchar_map):
            raise IndexError('Angular momentum {} out of range. Must be less than {}'.format(a, len(amchar_map)))
        amchar.append(amchar_map[a])

    return ''.join(amchar)


def amchar_to_int(amchar, hij=False):
    '''Convert an angular momentum integer to a character

    The return value is a list of integers (to handle sp, spd, ... orbitals)

    For example, converts 'p' to [1] and 'sp' to [0,1]

    If hij is True, the ordering spdfghijkl is used. Otherwise, the
    ordering will be spdfghikl (skipping j)
    '''

    if hij:
        amchar_map = _amchar_map_hij
    else:
        amchar_map = _amchar_map_hik

    amchar_lower = amchar.lower()

    amint = []

    for c in amchar_lower:
        if c not in amchar_map:
            raise KeyError('Angular momentum character {} is not valid'.format(c))

        amint.append(amchar_map.index(c))

    return amint


def electron_shells_start(nelectrons, max_am=20):
    '''Return the starting principle quantum numbers of electron shells

    For example, an ECP covering 10 electrons will covers 1s, 2s, 2p shells. The
    electrons shells will then start at 3s, 3p, 3d, and 4f (returned as [3, 3, 3, 4])

    If an ECP covers 30 electrons, then the shells will start at [5, 4, 4, 4]

    Only fully-covered shells are counted. If a shell is partly covered, an exception
    is raised.

    The returned list is extended up to max_am.

    Note: Since the main use of this is for ECPs, we only cover what can really be found
    on the periodic table. No excited states!

    Parameters
    ----------
    nelectrons : int
        Number of electrons covered by an ECP
    max_am : int
        Fill out the starting principal quantum numbers up to this am

    Returns
    -------
    list
        The starting principal quantum numbers of s, p, d, and f shells.
    '''

    if nelectrons < 0:
        raise RuntimeError("Can't have a negative number of electrons")
    if nelectrons > 118:
        raise NotImplementedError("Too many electrons for electrons_shells_start")

    # yapf: disable
    # The usual filling order of electrons you learned in high school
    # Tuple of (am, nelec)
    aminfo = ( (0, 2),                           # He
               (0, 2), (1, 6),                   # Ne
               (0, 2), (1, 6),                   # Ar
               (0, 2), (2, 10), (1, 6),          # Kr
               (0, 2), (2, 10), (1, 6),          # Xe
               (0, 2), (3, 14), (2, 10), (1, 6), # Rn
               (0, 2), (3, 14), (2, 10), (1, 6)  # Og
             )

    # But the above doesn't always work - ECPs don't always
    # cover things in such a regular order. So handle some
    # special ones
    special_am = { 28: [0, 0, 0, 1, 1, 2],
                   46: [0, 0, 0, 0, 1, 1, 1, 2, 2],
                   60: [0, 0, 0, 0, 1, 1, 1, 2, 2, 3],
                   68: [0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3],
                   78: [0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3],
                   92: [0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3]
                 }
    # yapf: enable

    if nelectrons in special_am:
        contained_am = special_am[nelectrons]
    else:
        contained_am = []

        for am, n in aminfo:
            if nelectrons >= n:
                nelectrons -= n
                contained_am.append(am)
            else:
                break

        if nelectrons != 0:
            raise RuntimeError("Electrons cover a partial shell. {} electrons left".format(nelectrons))

    start = [
        contained_am.count(0) + 1,
        contained_am.count(1) + 2,
        contained_am.count(2) + 3,
        contained_am.count(3) + 4
    ]
    start.extend(range(5, max_am + 2))

    return start



'''
Conversion of basis sets to Dalton format
'''

from .. import lut, manip, sort, misc, printing


def write_dalton(basis):
    '''Converts a basis set to Dalton format
    '''

    s = '! Basis = {}\n\n'.format(basis['name'])

    basis = manip.make_general(basis, False, True)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        for z in electron_elements:
            data = basis['elements'][z]
            #sym = lut.element_sym_from_Z(z, True)
            elname = lut.element_name_from_Z(z).upper()
            cont_string = misc.contraction_string(data)

            s += 'a {}\n'.format(z)
            s += '! {}       {}\n'.format(elname, cont_string)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 1
                nprim = len(exponents)
                ngen = len(coefficients)

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True)
                s += '! {} functions\n'.format(amchar)

                # Is this a bug in the original EMSL?
                #s += '{}   {}   1.00\n'.format(sym, r, nprim)
                s += '{}    {}    {}\n'.format('H', nprim, ngen)

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places, convert_exp=False)

    # Write out ECP
    if ecp_elements:
        s += '\n\nECP\n'
        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, normalize=True)
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{} nelec {}\n'.format(sym, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).upper()

                if am[0] == max_ecp_am:
                    s += '{} ul\n'.format(sym)
                else:
                    s += '{} {}\n'.format(sym, amchar)

                point_places = [0, 9, 32]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places, convert_exp=False)

        s += 'END\n'
    return s



'''
Conversion of basis sets to Turbomole format
'''

from .. import lut, manip, sort, printing


def write_turbomole(basis):
    '''Converts a basis set to Gaussian format
    '''

    # The role of the basis set determines what is put here
    role = basis.get('role', 'orbital')

    # By default, we will just use '$basis', unless otherwise specified
    s = '$basis\n'
    if role == 'jfit':
        s = '$jbas\n'
    elif role == 'jkfit':
        s = '$jkbas\n'
    elif role == 'rifit':
        s = '$cbas\n'

    s += '*\n'

    # TM basis sets are completely uncontracted
    basis = manip.uncontract_general(basis, True)
    basis = manip.uncontract_spdf(basis, 0, False)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        for z in electron_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, False)
            s += '{} {}\n'.format(sym, basis['name'])
            s += '*\n'

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 1
                nprim = len(exponents)

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am, hij=False)
                s += '    {}   {}\n'.format(nprim, amchar)

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places, convert_exp=True)

            s += '*\n'

    # Write out ECP
    if ecp_elements:
        s += '$ecp\n'
        s += '*\n'
        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z)
            s += '{} {}-ecp\n'.format(sym, basis['name'])
            s += '*\n'

            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])
            max_ecp_amchar = lut.amint_to_char([max_ecp_am], hij=True)

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '  ncore = {}   lmax = {}\n'.format(data['ecp_electrons'], max_ecp_am)

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True)

                if am[0] == max_ecp_am:
                    s += '{}\n'.format(amchar)
                else:
                    s += '{}-{}\n'.format(amchar, max_ecp_amchar)

                point_places = [9, 23, 32]
                s += printing.write_matrix([*coefficients, rexponents, gexponents], point_places, convert_exp=True)
            s += '*\n'

    s += '$end\n'
    return s



import re
from .. import lut
from . import helpers
from .nwchem import _parse_ecp_lines

all_element_names = [x.lower() for x in lut.all_element_names()]
all_element_names.append('wolffram')  # For tungsten

# Lines beginning a shell can be:
#   H {nprim} {ngen}
# or
#   {nprim} {ngen} 0
# This regex handles both cases
shell_begin_re = re.compile(r'^(?:[hH]\s+)?(\d+)\s+(\d+)(?: +0)?$')

# Typical format for starting an element with a comment
# Not sure this is absolutely required though
element_begin_re = re.compile(r'^!\s+([a-z]+)\s+\(.*\)\s*->\s*\[.*\]$')


#############################################################################
# There seems to be several dalton formats. One splits out by element, with
# each element block starting with "a {element_Z}". Then each shell
# starts with "{nprim} {ngen} 0"
# The second is also split by elements, but the element name
# is in a comment...
#############################################################################
def _line_begins_element(line):
    if not line:
        return False

    line = line.lower()

    if line.startswith('a '):
        return True

    if element_begin_re.match(line):
        s = line[1:].split()
        if s[0] in all_element_names:
            return True
        else:
            raise RuntimeError("Line looks to start an element, but element name is unknown to me. Line: " + line)

    return False


def _parse_electron_lines(basis_lines, bs_data):
    '''Parses lines representing all the electron shells for all elements

    Resulting information is stored in bs_data
    '''

    # A little bit of a hack here
    # If we find the start of an element, remove all the following comment lines
    new_basis_lines = []
    i = 0
    while i < len(basis_lines):
        if _line_begins_element(basis_lines[i]):
            new_basis_lines.append(basis_lines[i])
            i += 1
            while basis_lines[i].startswith('!'):
                i += 1
        else:
            new_basis_lines.append(basis_lines[i])
            i += 1

    basis_lines = new_basis_lines

    # Now split out all the element blocks
    element_blocks = helpers.partition_lines(basis_lines, _line_begins_element, min_size=3)

    # For each block, split out all the shells
    for el_lines in element_blocks:
        # Figure out which type of block this is (does it start with 'a ' or a comment
        header = el_lines[0].lower()
        if header.startswith('a '):
            element_Z = helpers.parse_line_regex(r'a +(\d+) *$', header, "a {element_z}", convert_int=False)
        elif header.startswith('!'):
            element_name = helpers.parse_line_regex(element_begin_re, header, '! {element_name}')
            element_Z = lut.element_Z_from_name(element_name, as_str=True)
        else:
            raise RuntimeError("Unable to parse block in dalton: header line is \"{}\"".format(header))

        element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells')
        el_lines.pop(0)

        # Remove all the rest of the comment lines
        el_lines = helpers.prune_lines(el_lines, '!')

        # Now partition again into blocks of shells for this element
        shell_blocks = helpers.partition_lines(el_lines, lambda x: shell_begin_re.match(x))

        # Shells are written in increasing angular momentum
        shell_am = 0

        for sh_lines in shell_blocks:
            nprim, ngen = helpers.parse_line_regex(shell_begin_re, sh_lines[0], 'nprim, ngen')
            exponents, coefficients = helpers.parse_primitive_matrix(sh_lines[1:], nprim=nprim, ngen=ngen)

            func_type = helpers.function_type_from_am([shell_am], 'gto', 'spherical')

            shell = {
                'function_type': func_type,
                'region': '',
                'angular_momentum': [shell_am],
                'exponents': exponents,
                'coefficients': coefficients
            }

            element_data['electron_shells'].append(shell)
            shell_am += 1


def read_dalton(basis_lines):
    '''Reads Dalton-formatted file data and converts it to a dictionary with the
       usual BSE fields
    '''

    ###########################################################################
    # We need to leave in comments until later, since they can be significant
    # (one format allows "! {ELEMENT}" to start an element block)
    ###########################################################################

    # But we still prune blank lines
    basis_lines = helpers.prune_lines(basis_lines)

    bs_data = {}

    # Skip forward until either:
    # 1. Line begins with 'a'
    # 2. Line begins with 'ecp'
    # 2. Lines begins with '!', with an element name following
    while basis_lines and not _line_begins_element(basis_lines[0]) and basis_lines[0].lower() != 'ecp':
        basis_lines.pop(0)

    # Empty file?
    if not basis_lines:
        return bs_data

    # Partition into ECP and electron blocks
    # I don't think Dalton supports ECPs, but the original BSE
    # Used the NWChem output format for the ECP part
    basis_sections = helpers.partition_lines(basis_lines, lambda x: x.lower() == 'ecp', min_blocks=1, max_blocks=2)

    for s in basis_sections:
        if s[0].lower() == 'ecp':
            _parse_ecp_lines(s, bs_data)
        else:
            _parse_electron_lines(s, bs_data)

    return bs_data



#!/usr/bin/env python3

import sys
import basis_set_exchange as bse

for f in sys.argv[1:]:
    a = bse.fileio.read_json_basis(f)
    a = bse.manip.make_general(a)
    a = bse.manip.prune_basis(a)
    bse.fileio.write_json_basis(f, a)



import re
from .. import lut
from . import helpers

am_line_re = re.compile(r'^([A-Za-z]+)\s+([A-Za-z]+)$')
nelec_re = re.compile(r'^([a-z]+)\s+nelec\s+(\d+)$', flags=re.IGNORECASE)


def _parse_electron_lines(basis_lines, bs_data):
    ''' Parses lines representing all the electron shells for all elements

    Resulting information is stored in bs_data
    '''

    # Remove 'end' from the lines (if they exist)
    # They may exist when this is called from other readers
    basis_lines = [x for x in basis_lines if x.lower() != 'end']

    # Start at index 1 in order to strip of the first line ('BASIS AO PRINT' or something)
    shell_blocks = helpers.partition_lines(basis_lines[1:], lambda x: x[0].isalpha(), min_size=2)

    for sh_lines in shell_blocks:
        element_sym, shell_am = helpers.parse_line_regex(am_line_re, sh_lines[0], "Element sym, shell am")
        shell_am = lut.amchar_to_int(shell_am)

        element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
        element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells', key_exist_ok=True)

        func_type = helpers.function_type_from_am(shell_am, 'gto', 'spherical')

        # How many columns of coefficients do we have?
        # Only if this is a fused shell do we know
        ngen = len(shell_am) if len(shell_am) > 1 else None

        exponents, coefficients = helpers.parse_primitive_matrix(sh_lines[1:], ngen=ngen)

        shell = {
            'function_type': func_type,
            'region': '',
            'angular_momentum': shell_am,
            'exponents': exponents,
            'coefficients': coefficients
        }

        element_data['electron_shells'].append(shell)


def _parse_ecp_lines(basis_lines, bs_data):
    ''' Parses lines representing all the ECP potentials for all elements

    Resulting information is stored in bs_data
    '''

    # Remove 'end' from the lines (if they exist)
    # They may exist when this is called from other readers
    basis_lines = [x for x in basis_lines if x.lower() != 'end']

    # Start at index 1 in order to strip of the first line ('ECP' or something)
    # This splits out based on starting with an alpha character. A block can be either
    #   a potential or the nelec line
    ecp_blocks = helpers.partition_lines(basis_lines[1:], lambda x: x[0].isalpha())

    for pot_lines in ecp_blocks:
        # Check if this is the nelec line
        if len(pot_lines) == 1:
            # Check for this and give a better error message
            if not nelec_re.match(pot_lines[0]):
                raise RuntimeError("Unknown block with single line. Perhaps it's an empty block? Line: " +
                                   pot_lines[0])

            element_sym, n_elec = helpers.parse_line_regex(nelec_re, pot_lines[0].lower(), "ECP: Element sym, nelec")

            element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
            element_data = helpers.create_element_data(bs_data, element_Z, 'ecp_electrons', create=int)
            element_data['ecp_electrons'] = n_elec
        else:
            element_sym, pot_am = helpers.parse_line_regex(am_line_re, pot_lines[0], "ECP: Element sym, pot AM")

            element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
            element_data = helpers.create_element_data(bs_data, element_Z, 'ecp_potentials', key_exist_ok=True)

            # See if this is the 'ul' angular momentum
            if pot_am.lower() == 'ul':
                pot_am = []  # Placeholder - leave for later
            else:
                pot_am = lut.amchar_to_int(pot_am)

            ecp_data = helpers.parse_ecp_table(pot_lines[1:])
            ecp_pot = {
                'angular_momentum': pot_am,
                'ecp_type': 'scalar_ecp',
                'r_exponents': ecp_data['r_exp'],
                'gaussian_exponents': ecp_data['g_exp'],
                'coefficients': ecp_data['coeff']
            }

            element_data['ecp_potentials'].append(ecp_pot)

    # Fix ecp angular momentum now that everything has been read
    # Specifically, we can set the 'ul' potential to be the max am + 1
    for el, v in bs_data.items():
        if 'ecp_potentials' not in v:
            continue

        all_ecp_am = []
        for x in v['ecp_potentials']:
            all_ecp_am.extend(x['angular_momentum'])

        max_ecp_am = max(all_ecp_am)

        for s in v['ecp_potentials']:
            if not s['angular_momentum']:
                s['angular_momentum'] = [max_ecp_am + 1]

    # Make sure the number of electrons replaced by the ECP was specified for all elements
    for el, v in bs_data.items():
        if 'ecp_potentials' in v and 'ecp_electrons' not in v:
            raise RuntimeError("Number of ECP electrons not specified for element {}".format(el))


def read_nwchem(basis_lines):
    '''Reads NWChem-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the nwchem format does not store all the fields we
       have, so some fields are left blank
    '''

    basis_lines = helpers.prune_lines(basis_lines, '#')

    bs_data = {}

    # split into basis and ecp
    basis_sections = helpers.partition_lines(basis_lines,
                                             lambda x: x.lower() == 'end',
                                             min_blocks=1,
                                             max_blocks=2,
                                             include_match=False)

    for s in basis_sections:
        if s[0].lower().startswith('basis'):
            _parse_electron_lines(s, bs_data)
        elif s[0].lower().startswith('ecp'):
            _parse_ecp_lines(s, bs_data)
        else:
            raise RuntimeError("Unknown section: " + s[0])

    return bs_data



'''
Miscellaneous helper functions
'''

import re
from . import lut


def _Z_from_str(s):
    if s.isdecimal():
        return int(s)
    else:
        return lut.element_Z_from_sym(s)


def transpose_matrix(mat):
    '''Transposes a matrix (list of lists) commonly done do coefficients'''

    return list(map(list, zip(*mat)))


def max_am(shells):
    '''Determine the maximum angular momentum of a list of shells or potentials'''
    all_am = [max(x['angular_momentum']) for x in shells]
    return max(all_am)


def contraction_string(element):
    """
    Forms a string specifying the contractions for an element

    ie, (16s,10p) -> [4s,3p]
    """

    # Does not have electron shells (ECP only?)
    if 'electron_shells' not in element:
        return ""

    cont_map = dict()
    for sh in element['electron_shells']:
        nprim = len(sh['exponents'])
        ngeneral = len(sh['coefficients'])

        # is a combined general contraction (sp, spd, etc)
        is_spdf = len(sh['angular_momentum']) > 1

        for am in sh['angular_momentum']:
            # If this a general contraction (and not combined am), then use that
            ncont = ngeneral if not is_spdf else 1

            if am not in cont_map:
                cont_map[am] = (nprim, ncont)
            else:
                cont_map[am] = (cont_map[am][0] + nprim, cont_map[am][1] + ncont)

    primstr = ""
    contstr = ""
    for am in sorted(cont_map.keys()):
        nprim, ncont = cont_map[am]

        if am != 0:
            primstr += ','
            contstr += ','
        primstr += str(nprim) + lut.amint_to_char([am])
        contstr += str(ncont) + lut.amint_to_char([am])

    return "({}) -> [{}]".format(primstr, contstr)


def compact_elements(elements):
    """
    Create a string (with ranges) given a list of element numbers

    For example, [1, 2, 3, 6, 7, 8, 10] will return "H-Li,C-O,Ne"
   """

    if not elements:
        return

    # We have to convert to integers for this function
    elements = [int(el) for el in elements]

    # Just to be safe, sort the list
    el = sorted(set(elements))

    ranges = []
    i = 0
    while i < len(el):
        start_el = el[i]
        end_el = start_el

        i += 1
        while i < len(el):
            if el[i] != end_el + 1:
                break

            end_el += 1
            i += 1

        if start_el == end_el:
            ranges.append([start_el])
        else:
            ranges.append([start_el, end_el])

    # Convert to elemental symbols
    range_strs = []
    for r in ranges:
        sym = lut.element_sym_from_Z(r[0], True)

        if len(r) == 1:
            range_strs.append(sym)
        elif len(r) == 2 and r[1] == r[0] + 1:
            sym2 = lut.element_sym_from_Z(r[1], True)
            range_strs.append(sym + "," + sym2)
        else:
            sym2 = lut.element_sym_from_Z(r[1], True)
            range_strs.append(sym + "-" + sym2)

    return ",".join(range_strs)


def expand_elements(compact_el, as_str=False):
    """
    Create a list of integers given a string or list of compacted elements

    This is partly the opposite of compact_elements, but is more flexible.

    compact_el can be a list or a string. If compact_el is a list, each element is processed individually
    as a string (meaning list elements can contain commas, ranges, etc)
    If compact_el is a string, it is split by commas and then each section is processed.

    In all cases, element symbols (case insensitive) and Z numbers (as integers or strings)
    can be used interchangeably. Ranges are also allowed in both lists and strings.

    Some examples:
        "H-Li,C-O,Ne" will return [1, 2, 3, 6, 7, 8, 10]
        "H-N,8,Na-12" will return [1, 2, 3, 4, 5, 6, 7, 8, 11, 12]
        ['C', 'Al-15,S', 17, '18'] will return [6, 13, 14, 15, 16, 17, 18]

    If as_str is True, the list will contain strings of the integers
    (ie, the first example above will return ['1', '2', '3', '6', '7', '8', '10']
    """

    # If an integer, just return it
    if isinstance(compact_el, int):
        if as_str:
            return [str(compact_el)]
        else:
            return [compact_el]

    # If compact_el is a list, make it a comma-separated string
    if isinstance(compact_el, list):
        compact_el = [str(x) for x in compact_el]
        compact_el = [x for x in compact_el if x]
        compact_el = ','.join(compact_el)

    # Find multiple - or ,
    # Also replace all whitespace with spaces
    compact_el = re.sub(r',+', ',', compact_el)
    compact_el = re.sub(r'-+', '-', compact_el)
    compact_el = re.sub(r'\s+', '', compact_el)

    # Find starting with or ending with comma and strip them
    compact_el = compact_el.strip(',')

    # Check if I was passed an empty string or list
    if not compact_el:
        return []

    # Find some erroneous patterns
    # -, and ,-
    if '-,' in compact_el:
        raise RuntimeError("Malformed element string")
    if ',-' in compact_el:
        raise RuntimeError("Malformed element string")

    # Strings ends or begins with -
    if compact_el.startswith('-') or compact_el.endswith('-'):
        raise RuntimeError("Malformed element string")

    # x-y-z
    if re.search(r'\w+-\w+-\w+', compact_el):
        raise RuntimeError("Malformed element string")

    # Split on commas
    tmp_list = compact_el.split(',')

    # Now go over each one and replace elements with ints
    el_list = []
    for el in tmp_list:
        if '-' not in el:
            el_list.append(_Z_from_str(el))
        else:
            begin, end = el.split('-')
            begin = _Z_from_str(begin)
            end = _Z_from_str(end)
            el_list.extend(list(range(begin, end + 1)))

    if as_str:
        return [str(x) for x in el_list]
    else:
        return el_list


def transform_basis_name(name):
    """
    Transforms the name of a basis set to an internal representation

    This makes comparison of basis set names easier by, for example,
    converting the name to all lower case.
    """

    name = name.lower()
    name = name.replace('/', '_sl_')
    name = name.replace('*', '_st_')
    return name


def basis_name_to_filename(name):
    '''
    Given a basis set name, transform it into a valid filename

    This makes sure filenames don't contain invalid characters
    '''

    return transform_basis_name(name)


def basis_name_from_filename(filename):
    '''
    Given a basis set name that was part of a filename, determine the basis set name

    This is opposite of :func:`transform_basis_name`

    Pass only the part of the filename that contains the basis set name
    '''

    name = filename.lower()
    name = name.replace('_sl_', '/')
    name = name.replace('_st_', '*')
    return name



import re
from .. import lut, misc
from . import helpers

element_head_re = re.compile(r'^/([a-zA-Z]{1,3})\.(?:ECP\.)?([^.]+)\..*$')
electron_head_re = re.compile(r'^/([a-zA-Z]{1,3})\.([^.]+)\..*$')
ecp_head_re = re.compile(r'^/([a-zA-Z]{1,3})\.ECP\.([^.]+)\..*$')

electron_z_max_am_re = re.compile(r'^(\d+|{})(?:\s+(\d+))?$'.format(helpers.floating_re_str))
shell_nprim_ngen_re = re.compile(r'^(\d+)(?:\s+(\d+))?$')

ecp_info_re = re.compile(r'^[Pp]{2}\s*,\s*([a-zA-Z]+)\s*,\s*(\d+)\s*,\s*(\d+)\s*;$')
ecp_pot_begin_re = re.compile(r'^(\d+)\s*;.*$')  # Sometime comments are after the semicolon


def _parse_electron_lines(basis_lines, bs_data, element_Z):
    element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells')

    # Handle the options block
    # This specifies the kind of data that might be found at the end of the element block
    # We ignore that data, but we need to know it is there
    options_lines, basis_lines = helpers.remove_block(basis_lines, 'Options', 'EndOptions')

    # Every option adds another block after each shell.
    # I don't care about the contents, just how many options there are
    n_option_blocks = len(options_lines)

    # Next is the nuclear charge and the max_am
    # max_am may be missing
    nuc_charge, max_am = helpers.parse_line_regex(electron_z_max_am_re, basis_lines[0], 'Electron: Z, max_am')

    # If the nuclear charge is not equal to the element Z, then this must be an ECP
    # If the number of ECP electrons already exists, check it.
    # If not, put it into the element data, and the ecp reader below will double check it
    nuc_charge = float(nuc_charge)

    # Is this actually an integer?
    if float(int(nuc_charge)) != nuc_charge:
        raise RuntimeError("Non-integer specified for nuclear charge: " + nuc_charge)

    ecp_electrons = int(element_Z) - int(nuc_charge)
    if 'ecp_electrons' in element_data and element_data['ecp_electrons'] != ecp_electrons:
        raise RuntimeError("Element Z: {} with charge {} does not match ECP electrons {}".format(
            element_Z, nuc_charge, ecp_electrons))
    elif ecp_electrons > 0:
        element_data['ecp_electrons'] = ecp_electrons

    # Partition the remaining lines. Blocks start with one or two integers.
    # This includes typical electron shell blocks, and blocks corresponding to options
    shell_blocks = helpers.partition_lines(basis_lines[1:], lambda x: x.split()[0].isdecimal())

    # There should be max_am + 1 shell blocks
    # Each shell will also have n_option_blocks additional blocks
    if max_am is not None and len(shell_blocks) != (max_am + 1) * (n_option_blocks + 1):
        raise RuntimeError("Found {} shell blocks. Expected {}".format(len(shell_blocks),
                                                                       (max_am + 1) * (n_option_blocks + 1)))

    # Shells are simply increasing AM
    shell_am = 0

    # Loop over all shell lines before the option block data
    # Option blocks are after each shell, so just slice and skip
    # every n_option_blocks
    if n_option_blocks > 0:
        shell_blocks = shell_blocks[::n_option_blocks + 1]

    for shell_lines in shell_blocks:
        nprim, ngen = helpers.parse_line_regex(shell_nprim_ngen_re, shell_lines[0], 'Shell nprim, ngen')

        if nprim <= 0:
            raise RuntimeError("Cannot have {} primitives in a shell".format(nprim))
        if ngen is not None and ngen <= 0:
            raise RuntimeError("Cannot have {} general contractions in a shell".format(ngen))

        exponents, shell_lines = helpers.read_n_floats(shell_lines[1:], nprim)

        # Read the coefficient matrix from the next block
        # We may or may not know the dimentions of the matrix since &@*!# is optional
        # in this format.
        coefficients = helpers.read_all_floats(shell_lines)

        # Now we can check if this makes sense
        n_coefs = len(coefficients)
        if n_coefs == 0:
            raise RuntimeError("Have zero coefficients?")
        if n_coefs % nprim != 0:
            raise RuntimeError("Number of coefficients is not a multiple of nprim: {} % {} = {}".format(
                n_coefs, nprim, n_coefs % nprim))

        # If we do actually have the number of general contractions, does it match?
        if ngen is not None and ngen != n_coefs // nprim:
            raise RuntimeError("Expected {} general contractions, but found {}".format(ngen, n_coefs // nprim))
        else:
            ngen = n_coefs // nprim

        # Turn the coefficients into a matrix
        coefficients = helpers.chunk_list(coefficients, nprim, ngen)
        coefficients = misc.transpose_matrix(coefficients)

        # Now add to the bs_data
        func_type = helpers.function_type_from_am([shell_am], 'gto', 'spherical')

        shell = {
            'function_type': func_type,
            'region': '',
            'angular_momentum': [shell_am],
            'exponents': exponents,
            'coefficients': coefficients
        }

        element_data['electron_shells'].append(shell)
        shell_am += 1


def _parse_ecp_lines(basis_lines, bs_data, element_Z):
    # Remove "Spectral" Stuff
    _, basis_lines = helpers.remove_block(basis_lines, r'^Spectral.*', r'^End\s*Of\s*Spectral.*')

    # Parse the ecp info line
    element_sym, ecp_electrons, max_am = helpers.parse_line_regex(ecp_info_re, basis_lines[0],
                                                                  "ECP Info: pp,sym,nelec,max_am")
    element_Z_ecp = lut.element_Z_from_sym(element_sym, as_str=True)

    # Does this block match the element symbol from the main element header?
    if element_Z_ecp != element_Z:
        raise RuntimeError("ECP element Z={} found in block for element Z={}".format(element_Z, element_Z_ecp))

    element_data = helpers.create_element_data(bs_data, element_Z, 'ecp_potentials')

    # Does the ecp_electrons key exist? This may have been determined when reading the
    # electron shells above
    if 'ecp_electrons' in element_data and element_data['ecp_electrons'] != ecp_electrons:
        raise RuntimeError(
            "No. of electrons specified in ECP block do not match already-determined number of electrons: {} vs {}".
            format(ecp_electrons, element_data['ecp_electrons']))
    else:
        element_data['ecp_electrons'] = ecp_electrons

    # Now split into potentials
    # The beginning of each potential is a number followed by a semicolon
    pot_blocks = helpers.partition_lines(basis_lines[1:], ecp_pot_begin_re.match, min_size=2)

    if len(pot_blocks) != max_am + 1:
        raise RuntimeError("Expected {} potentials, but got {}".format(max_am + 1, len(pot_blocks)))

    # Set up the AM for the potentials
    all_pot_am = helpers.potential_am_list(max_am)

    for pot_lines in pot_blocks:
        pot_am = all_pot_am.pop(0)

        nlines = helpers.parse_line_regex(ecp_pot_begin_re, pot_lines[0], "ECP Potential: # of lines")
        if nlines != len(pot_lines) - 1:
            raise RuntimeError("Expected {} lines in potential, but got {}".format(nlines, len(pot_lines) - 1))

        # Strip trailing semicolon
        pot_lines = [x.rstrip(';') for x in pot_lines[1:]]
        ecp_data = helpers.parse_ecp_table(pot_lines, split=r'\s*,\s*')
        ecp_pot = {
            'angular_momentum': [pot_am],
            'ecp_type': 'scalar_ecp',
            'r_exponents': ecp_data['r_exp'],
            'gaussian_exponents': ecp_data['g_exp'],
            'coefficients': ecp_data['coeff']
        }

        element_data['ecp_potentials'].append(ecp_pot)


def read_molcas(basis_lines):
    '''Reads molcas-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the molcas format does not store all the fields we
       have, so some fields are left blank
    '''

    basis_lines = helpers.prune_lines(basis_lines, '*#$')

    bs_data = {}

    # Split into elements. Every start of an element is /
    element_blocks = helpers.partition_lines(basis_lines, lambda x: x.startswith('/'), min_size=4)

    # Inside this loop, check that all blocks refer to the same basis set
    basis_names_found = set()

    for element_lines in element_blocks:
        # The start of the element block is:
        #     /{element_sym}.{bs_name}.stuff
        #     Reference line 1
        #     Reference line 2
        # The next line should either be "options" or the element Z number
        element_sym, basis_name = helpers.parse_line_regex(element_head_re, element_lines[0], 'Start of element line')
        element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
        basis_names_found.add(basis_name.lower())

        # remove the header and comments
        element_lines = element_lines[3:]

        # Split based on PP (pseudopotential)
        element_split = helpers.partition_lines(element_lines,
                                                lambda x: x.lower().startswith('pp,'),
                                                min_blocks=1,
                                                max_blocks=2)

        for block_lines in element_split:
            if block_lines[0].lower().startswith('pp'):
                _parse_ecp_lines(block_lines, bs_data, element_Z)
            else:
                _parse_electron_lines(block_lines, bs_data, element_Z)

    # Check for multiple basis sets
    if len(basis_names_found) > 1:
        raise RuntimeError("Multiple basis sets found in file: " + ','.join(basis_names_found))

    return bs_data



'''
Conversion of basis sets to NWChem format
'''

from .. import lut, manip, printing, misc, sort


def write_nwchem(basis):
    '''Converts a basis set to NWChem format
    '''

    # Uncontract all but SP
    basis = manip.uncontract_spdf(basis, 1, True)
    basis = sort.sort_basis(basis, False)

    s = ''

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    if electron_elements:
        # basis set starts with a string
        s += 'BASIS "ao basis" PRINT\n'

        # Electron Basis
        for z in electron_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, True)
            s += '#BASIS SET: {}\n'.format(misc.contraction_string(data))

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 1

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am).upper()
                s += '{}    {}\n'.format(sym, amchar)

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places)

        s += 'END\n'

    # Write out ECP
    if ecp_elements:
        s += '\n\nECP\n'

        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, True)
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{} nelec {}\n'.format(sym, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).upper()

                if am[0] == max_ecp_am:
                    s += '{} ul\n'.format(sym)
                else:
                    s += '{} {}\n'.format(sym, amchar)

                point_places = [0, 10, 33]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places)

        s += 'END\n'

    return s



'''
Sorting of BSE related dictionaries and data
'''

import sys
import copy

# Dictionaries for python 3.6 and above are insertion ordered
# For other versions, use an OrderedDict
_use_odict = sys.version_info.major == 3 and sys.version_info.minor < 6

if _use_odict:
    from collections import OrderedDict


def sort_basis_dict(bs):
    """Sorts a basis set dictionary into a standard order

    This, for example, allows the written file to be more easily read by humans by,
    for example, putting the name and description before more detailed fields.

    This is generally for cosmetic reasons. However, users will generally like things
    in a consistent order
    """

    # yapf: disable
    _keyorder = [
        # Schema stuff
        'molssi_bse_schema', 'schema_type', 'schema_version',

        # Auxiliary block
        'jkfit', 'jfit', 'rifit', 'optri', 'admmfit', 'dftxfit', 'dftjfit', 'guess',

        # Basis set metadata
        'name', 'names', 'aliases', 'tags', 'family', 'description', 'role', 'auxiliaries',
        'notes', 'function_types',

        # Reference stuff
        'reference_description', 'reference_keys',

        # Version metadata
        'version', 'revision_description', 'revision_date',

        # Sources of components
        'data_source',

        # Elements and data
        'elements', 'references', 'ecp_electrons',
        'electron_shells', 'ecp_potentials', 'components',

        # Shell information
        'function_type', 'region', 'angular_momentum', 'exponents',
        'coefficients',
        'ecp_type', 'angular_momentum', 'r_exponents', 'gaussian_exponents',
        'coefficients'
    ]
    # yapf: enable

    # Add integers for the elements (being optimistic that element 150 will be found someday)
    _keyorder.extend([str(x) for x in range(150)])

    bs_sorted = sorted(bs.items(), key=lambda x: _keyorder.index(x[0]))
    if _use_odict:
        bs_sorted = OrderedDict(bs_sorted)
    else:
        bs_sorted = dict(bs_sorted)

    for k, v in bs_sorted.items():
        # If this is a dictionary, sort recursively
        # If this is a list, sort each element but DO NOT sort the list itself.
        if isinstance(v, dict):
            bs_sorted[k] = sort_basis_dict(v)
        elif isinstance(v, list):
            # Note - the only nested list is with coeffs, which shouldn't be sorted
            #        (so we don't have to recurse into lists of lists)
            bs_sorted[k] = [sort_basis_dict(x) if isinstance(x, dict) else x for x in v]

    return bs_sorted


def sort_shell(shell, use_copy=True):
    """
    Sort a basis set shell into a standard order

    If use_copy is True, the input shells are not modified.
    """

    if use_copy:
        shell = copy.deepcopy(shell)

    tmp_c = shell['coefficients']

    # Sort columns by number of non-zero entries
    # But don't sort columns for SP shells
    if len(shell['angular_momentum']) == 1:
        col_counts = [sum([1 for x in y if float(x) != 0.0]) for y in tmp_c]
        tmp_c = sorted(zip(col_counts, tmp_c), key=lambda x: x[0], reverse=True)
        tmp_c = list(zip(*tmp_c))[1]

    # Transpose of coefficients
    tmp_c = list(map(list, zip(*tmp_c)))

    # For each primitive, find the index of the first nonzero coefficient
    nonzero_idx = [next((i for i, x in enumerate(c) if float(x) != 0.0), None) for c in tmp_c]

    # Zip together exponents and coeffs for sorting
    tmp = zip(shell['exponents'], tmp_c, nonzero_idx)

    # Sort by decreasing value of exponent
    tmp = sorted(tmp, key=lambda x: -float(x[0]))

    # Now (stable) sort by first non-zero coefficient
    tmp = sorted(tmp, key=lambda x: int(x[2]))

    # Unpack, and re-transpose the coefficients
    tmp_c = [x[1] for x in tmp]
    shell['exponents'] = [x[0] for x in tmp]

    tmp_c = list(map(list, zip(*tmp_c)))

    shell['coefficients'] = tmp_c

    return shell


def sort_shells(shells, use_copy=True):
    """
    Sort a list of basis set shells into a standard order

    The order within a shell is by decreasing value of the exponent.

    The order of the shell list is in increasing angular momentum, and then
    by decreasing number of primitives, then decreasing value of the largest exponent.

    If use_copy is True, the input shells are not modified.
    """

    if use_copy:
        shells = copy.deepcopy(shells)

    # Sort primitives within a shell
    # (copying already handled above)
    shells = [sort_shell(sh, False) for sh in shells]

    # Sort the list by increasing AM, then general contraction level, then decreasing highest exponent
    # yapf: disable
    return list(
        sorted(
            shells,
            key=lambda x: (max(x['angular_momentum']),
                          -len(x['exponents']),
                          -len(x['coefficients']),
                          -max(float(y) for y in x['exponents']))))
    # yapf: enable


def sort_potentials(potentials, use_copy=True):
    """
    Sort a list of ECP potentials into a standard order

    The order within a potential is not modified.

    The order of the shell list is in increasing angular momentum, with the largest
    angular momentum being moved to the front.

    If use_copy is True, the input potentials are not modified.
    """

    if use_copy:
        potentials = copy.deepcopy(potentials)

    # Sort by increasing AM, then move the last element to the front
    potentials = list(sorted(potentials, key=lambda x: x['angular_momentum']))
    potentials.insert(0, potentials.pop())
    return potentials


def sort_basis(basis, use_copy=True):
    """
    Sorts all the information in a basis set into a standard order

    If use_copy is True, the input basis set is not modified.
    """

    if use_copy:
        basis = copy.deepcopy(basis)

    for k, el in basis['elements'].items():
        if 'electron_shells' in el:
            el['electron_shells'] = sort_shells(el['electron_shells'], False)
        if 'ecp_potentials' in el:
            el['ecp_potentials'] = sort_potentials(el['ecp_potentials'], False)

    return sort_basis_dict(basis)


def sort_single_reference(ref_entry):
    """Sorts a dictionary containing data for a single reference into a standard order
    """

    # yapf: disable
    _keyorder = [
        # Schema stuff
        # This function gets called on the schema 'entry', too
        'schema_type', 'schema_version',

        # Type of the entry
        '_entry_type', 'type',

        # Actual publication info
        'authors', 'title', 'booktitle', 'series', 'editors', 'journal',
        'institution', 'school', 'volume', 'number', 'pages', 'year', 'note', 'publisher',
        'address', 'isbn', 'doi'
    ]
    # yapf: enable

    sorted_entry = sorted(ref_entry.items(), key=lambda x: _keyorder.index(x[0]))

    if _use_odict:
        return OrderedDict(sorted_entry)
    else:
        return dict(sorted_entry)


def sort_references_dict(refs):
    """Sorts a reference dictionary into a standard order

    The keys of the references are also sorted, and the keys for the data for each
    reference are put in a more canonical order.
    """

    if _use_odict:
        refs_sorted = OrderedDict()
    else:
        refs_sorted = dict()

    # We insert this first, That is ok - it will be overwritten
    # with the sorted version later
    refs_sorted['molssi_bse_schema'] = refs['molssi_bse_schema']

    # This sorts the entries by reference key (author1985a, etc)
    for k, v in sorted(refs.items()):
        refs_sorted[k] = sort_single_reference(v)

    return refs_sorted



'''
Some common formatting functions, etc
'''


def format_columns(lines, prefix=''):
    '''
    Create a simple column output

    Parameters
    ----------
    lines : list
        List of lines to format. Each line is a tuple/list with each
        element corresponding to a column
    prefix : str
        Characters to insert at the beginning of each line

    Returns
    -------
    str
        Columnated output as one big string
    '''
    if not lines:
        return ''

    ncols = 0
    for l in lines:
        ncols = max(ncols, len(l))

    if ncols == 0:
        return ''

    # We only find the max strlen for all but the last col
    maxlen = [0] * (ncols - 1)
    for l in lines:
        for c in range(ncols - 1):
            maxlen[c] = max(maxlen[c], len(l[c]))

    fmtstr = prefix + '  '.join(['{{:{x}}}'.format(x=x) for x in maxlen])
    fmtstr += '  {}'
    return [fmtstr.format(*l) for l in lines]



#!/usr/bin/env python3

import os
import tempfile
from .. import api, fileio
from ..misc import compact_elements

try:
    import graphviz
    graphviz_avail = True
except ImportError:
    graphviz_avail = False


def _make_graph(bsname, version=None, data_dir=None):
    '''
    Create a DOT graph file of the files included in a basis set
    '''

    if not graphviz_avail:
        raise RuntimeError("graphviz package is not installed")

    data_dir = api.fix_data_dir(data_dir)

    md = api._get_basis_metadata(bsname, data_dir)

    if version is None:
        version = md['latest_version']
    else:
        version = str(version)

    if version not in md['versions']:
        raise RuntimeError("Version {} of {} doesn't exist".format(version, bsname))

    gr = graphviz.Digraph(comment='Basis Set Graph: ' + bsname)

    # Read the table file
    table_path = os.path.join(data_dir, md['versions'][version]['file_relpath'])
    table_data = fileio.read_json_basis(table_path)

    table_edges = {}
    for el, entry in table_data['elements'].items():
        if entry not in table_edges:
            table_edges[entry] = []
        table_edges[entry].append(el)

    for k, v in table_edges.items():
        gr.edge(bsname, k, label=compact_elements(v))

    # Element file
    for elfile in table_edges.keys():
        element_path = os.path.join(data_dir, elfile)
        element_data = fileio.read_json_basis(element_path)

        element_edges = {}

        for el, components in element_data['elements'].items():
            components = components['components']
            components_str = '\n'.join(components)

            # skip if this element for the table basis doesn't come from this file
            if el not in table_data['elements']:
                continue
            if table_data['elements'][el] != elfile:
                continue

            if components_str not in element_edges:
                element_edges[components_str] = []
            element_edges[components_str].append(el)

        for k, v in element_edges.items():
            if len(v):
                gr.edge(elfile, k, label=compact_elements(v))

    return gr


def view_graph(bsname, version=None, data_dir=None):
    gr = _make_graph(bsname, version, data_dir)

    outdir = tempfile.mkdtemp()
    gr.render(directory=outdir, format='png', view=True)


def make_graph_file(bsname, outfile, render=False, version=None, data_dir=None):
    gr = _make_graph(bsname, version, data_dir)

    if render:
        gr.render(outfile, format='png', view=False)
    else:
        gr.save(outfile)



'''
Functions for basis set conversion
'''

from .readers import read_formatted_basis_file, read_formatted_basis_str
from .writers import write_formatted_basis_file, write_formatted_basis_str


def convert_formatted_basis_str(basis_in, in_fmt, out_fmt):
    '''Convert a formatted basis set to another format

    Parameters
    ----------
    basis_in : str
        String representing the formatted input basis set input
    in_fmt : str
        The format of the basis set stored in basis_in
    out_fmt : str
        The desired output format

    Returns
    -------
    str
        The basis set as a str with the new format
    '''

    basis_dict = read_formatted_basis_str(basis_in, in_fmt, validate=True, as_component=False)
    return write_formatted_basis_str(basis_dict, out_fmt)


def convert_formatted_basis_file(file_path_in, file_path_out, in_fmt=None, out_fmt=None, encoding='utf-8-sig'):
    '''Convert a formatted basis set file to another format

    Parameters
    ----------
    file_path_in : str
        Path to the file to be read
    file_path_out : str
        Path to the file to be written.
    in_fmt : str
        The format of the basis to be read. If None, it is detected from the file name
    out_fmt : str
        The format of the basis to be written. If None, it is detected from the file name
    encoding : str
        The encoding of the input file

    Returns
    -------
    str
        The basis set as a str with the new format
    '''

    basis_dict = read_formatted_basis_file(file_path_in,
                                           basis_fmt=in_fmt,
                                           encoding=encoding,
                                           validate=True,
                                           as_component=False)

    write_formatted_basis_file(basis_dict, file_path_out, basis_fmt=out_fmt)



'''
Conversion of basis sets to QCSchema JSON format
'''

import json
from .. import manip, sort, lut


def write_qcschema(basis):
    '''Converts a basis set to QCSchema JSON

    Note that the output is a string
    '''

    # Uncontract all but SP
    basis = manip.uncontract_spdf(basis, 1, True)
    basis = sort.sort_basis(basis, True)

    basis_name = basis.get('name', 'unknown_basis')
    basis_desc = basis.get('description', '<no description>')
    new_basis = {
        'schema_name': 'qcschema_basis',
        'schema_version': 1,
        'name': basis_name,
        'description': basis_desc
    }

    # For the 'center_data' key in the schema
    center_data = {}
    for el, eldata in basis['elements'].items():
        entry_name = lut.element_sym_from_Z(el) + '_' + basis_name

        eldata.pop('references', None)

        if 'electron_shells' in eldata:
            # Fix up some naming
            for sh in eldata['electron_shells']:
                sh.pop('region')
                func = sh.pop('function_type')
                if func == 'gto_spherical':
                    sh['harmonic_type'] = 'spherical'
                else:
                    # Set to cartesian if explicitely cartesian, or if it is an
                    # s or p shell
                    sh['harmonic_type'] = 'cartesian'

        if 'ecp_electrons' in eldata:
            for pot in eldata['ecp_potentials']:
                pot['ecp_type'] = pot['ecp_type'].replace('_ecp', '')

        center_data[entry_name] = eldata

    new_basis['center_data'] = center_data

    return json.dumps(new_basis, indent=2, ensure_ascii=False)



"""
Tests BSE curation functions
"""

import os
import pytest

from basis_set_exchange import printing, fileio, api
from .common_testvars import data_dir

# yapf: disable
@pytest.mark.parametrize('basis, element', [
                              ['6-31g', '8'],
                              ['CRENBL', '3'],
                              ['CRENBL', '92'],
                              ['LANL2DZ', '78']
                         ])
# yapf: enable
def test_print_element_data(basis, element):
    eldata = api.get_basis(basis)['elements'][element]
    printing.element_data_str(element, eldata)


# yapf: disable
@pytest.mark.parametrize('file_path', [
                              'dunning/cc-pVDZ.1.json',
                              'crenb/CRENBL.0.json',
                              'crenb/CRENBL-ECP.0.json'
                         ])
# yapf: enable
def test_print_component_basis(file_path):
    full_path = os.path.join(data_dir, file_path)
    comp = fileio.read_json_basis(full_path)
    printing.component_basis_str(comp)



'''                                                                                                      
Conversion of references to ris format                                                                                              
'''


def write_ris(key, ref):
    '''Convert a single reference to ris format                                                                                  
    '''
    s = ''

    s += '  \n'.format(ref['_entry_type'], key)
    if ref['_entry_type'] == 'article':
        s += "TY Journal Article \n"
    elif ref['_entry_type'] == 'misc':
        s += "TY Generic \n"
    elif ref['_entry_type'] == 'unpublished':
        s += "TY Unpublished \n"
    elif ref['_entry_type'] == 'incollection':
        s += "TY Book \n"
    elif ref['_entry_type'] == 'phdthesis':
        s += "TY Thesis \n"
    elif ref['_entry_type'] == 'techreport':
        s += "TY Report \n"
    else:
        s += "TY Generic\n"

    entry_lines = []
    for k, v in ref.items():
        if k == '_entry_type':
            continue

        # Handle authors/editors
        if k == 'authors':
            for author in v:
                entry_lines.append('AU {}'.format(author))
        elif k == 'year':
            entry_lines.append('PY {}'.format(v))
        elif k == 'journal':
            entry_lines.append('JO {}'.format(v))
        elif k == 'volume':
            entry_lines.append('VL {}'.format(v))
        elif k == 'pages':
            entry_lines.append('SP {}'.format(v))
        elif k == 'title':
            entry_lines.append('T1 {}'.format(v))
        elif k == 'doi':
            entry_lines.append('DO {}'.format(v))
        else:
            entry_lines.append('N1 {}:{}'.format(k, v))

    s += '\n'.join(entry_lines)
    s += '\n'

    return s



'''
Helpers for handling BSE metadata
'''

import os

from ..compose import compose_table_basis
from ..fileio import get_all_filelist, read_json_basis, _write_plain_json
from ..misc import transform_basis_name


def create_metadata_file(output_path, data_dir):
    '''Creates a METADATA.json file from a data directory

    The file is written to output_path
    '''

    # Relative path to all (BASIS).metadata.json files
    meta_filelist, table_filelist, _, _ = get_all_filelist(data_dir)

    metadata = {}
    for meta_file_relpath in meta_filelist:

        # Read in the metadata for a single basis set
        meta_file_path = os.path.join(data_dir, meta_file_relpath)
        bs_metadata = read_json_basis(meta_file_path)

        # Base of the filename for table basis sets
        # Basename is something like '6-31G.', including the last period
        base_relpath, meta_filename = os.path.split(meta_file_relpath)
        base_filename = meta_filename.split('.')[0] + '.'

        # All the table files that correspond to this metadata file
        # (relative to data_dir)

        this_filelist = [
            x for x in table_filelist
            if os.path.dirname(x) == base_relpath and os.path.basename(x).startswith(base_filename)
        ]

        # The 'versions' dict that will go into the metadata
        version_info = {}

        # Make sure function types are the same
        function_types = None

        # For each table basis, compose it
        for table_file in this_filelist:
            # Obtain just the filename of the table basis
            table_filename = os.path.basename(table_file)

            # Obtain the base filename and version from the filename
            # The base filename is the part before the first period
            # (filebase.ver.table.json)
            table_filebase, ver, _, _ = table_filename.split('.')

            # Fully compose the basis set from components
            bs = compose_table_basis(table_file, data_dir)

            # Elements for which this basis is defined
            defined_elements = sorted(list(bs['elements'].keys()), key=lambda x: int(x))

            # Determine the types of functions contained in the basis
            # (gto, ecp, etc)
            if function_types is None:
                function_types = bs['function_types']
            elif function_types != bs['function_types']:
                raise RuntimeError("Differing function types across versions for " + base_filename)

            # Create the metadata for this specific version
            # yapf: disable
            version_info[ver] = { 'file_relpath': table_file,
                                  'revdesc': bs['revision_description'],
                                  'revdate': bs['revision_date'],
                                  'elements': defined_elements
                                }
            # yapf: enable

        # Sort the version dicts
        version_info = dict(sorted(version_info.items()))
        # Find the maximum version for this basis
        latest_ver = max(version_info.keys())

        # Create the common metadata for this basis set
        # display_name and other_names are placeholders to keep order
        # yapf: disable
        common_md = { 'display_name': None,
                      'other_names': None,
                      'description': bs['description'],
                      'latest_version': latest_ver,
                      'tags': bs['tags'],
                      'basename': base_filename[:-1], # Strip off that trailing period
                      'relpath': base_relpath,
                      'family': bs['family'],
                      'role': bs['role'],
                      'function_types': function_types,
                      'auxiliaries': bs['auxiliaries'],
                      'versions': version_info }
        # yapf: enable

        # Loop through all the common names, translate them, and then add the data
        for bs_name in bs_metadata['names']:
            tr_name = transform_basis_name(bs_name)

            if tr_name in metadata:
                raise RuntimeError("Duplicate basis set name: " + tr_name)

            # Create a new entry, with all the common metadata
            # Also, store the other names for this basis
            other_names = bs_metadata['names'].copy()
            other_names.remove(bs_name)
            metadata[tr_name] = common_md.copy()
            metadata[tr_name]['display_name'] = bs_name
            metadata[tr_name]['other_names'] = other_names

    # Write out the metadata
    metadata = dict(sorted(metadata.items()))
    _write_plain_json(output_path, metadata)



import re
from .. import lut
from . import helpers

section_re = re.compile(r'^\$(basis|ecp|cbas|jbas|jkbas)$')
element_re = re.compile(r'^([a-zA-Z]{1,3})\s+(.*)$')
shell_re = re.compile(r'^(\d+) +([a-zA-Z])$')
ecp_info_re = re.compile(r'^ncore\s*=\s*(\d+)\s+lmax\s*=\s*(\d+)$', flags=re.IGNORECASE)
ecp_pot_am_re = re.compile(r'^([a-z])(-[a-z])?$')


def _parse_electron_lines(basis_lines, bs_data):
    # Strip all lines beginning with $
    basis_lines = helpers.prune_lines(basis_lines, '$')

    # Last line should be *
    # We don't need it
    if basis_lines[-1] != '*':
        raise RuntimeError("Missing terminating * line")
    basis_lines.pop()

    # Partition based on lines beginning with a character
    element_blocks = helpers.partition_lines(basis_lines, element_re.match, before=1, min_size=4)

    # Element lines should be surrounded by *
    # Check all first. the partition_lines above will eat part of the previous
    # element if the * is missing
    for element_lines in element_blocks:
        if element_lines[0] != '*':
            raise RuntimeError("Element line not preceded by *")
        if element_lines[2] != '*':
            raise RuntimeError("Element line not followed by *")

        # Check for any other lines starting with *
        for line in element_lines[3:]:
            if line.startswith('*'):
                raise RuntimeError("Found line starting with * that probably doesn't belong: " + line)

    # Now process them all
    for element_lines in element_blocks:
        element_sym, _ = helpers.parse_line_regex(element_re, element_lines[1], 'Element line')

        element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
        element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells')

        # Partition into shells
        shell_blocks = helpers.partition_lines(element_lines[3:], shell_re.match, min_size=2)

        for sh_lines in shell_blocks:
            nprim, shell_am = helpers.parse_line_regex(shell_re, sh_lines[0], 'shell nprim, am')
            shell_am = lut.amchar_to_int(shell_am)

            func_type = helpers.function_type_from_am(shell_am, 'gto', 'spherical')

            exponents, coefficients = helpers.parse_primitive_matrix(sh_lines[1:], nprim=nprim, ngen=1)

            shell = {
                'function_type': func_type,
                'region': '',
                'angular_momentum': shell_am,
                'exponents': exponents,
                'coefficients': coefficients
            }

            element_data['electron_shells'].append(shell)


def _parse_ecp_potential_lines(element_lines, bs_data):
    #########################################################
    # This is split out because the turbomole ECP format is
    # almost identical to the genbas ECP format
    #########################################################
    element_sym, _ = helpers.parse_line_regex(element_re, element_lines[0], 'Element line')

    element_Z = lut.element_Z_from_sym(element_sym, as_str=True)

    # We don't need the return value - we will use the one from creating ecp_electrons
    helpers.create_element_data(bs_data, element_Z, 'ecp_potentials')

    # 4th line should be ncore and lmax
    n_elec, max_am = helpers.parse_line_regex(ecp_info_re, element_lines[1], 'ECP ncore, lmax')

    element_data = helpers.create_element_data(bs_data, element_Z, 'ecp_electrons', key_exist_ok=False, create=int)
    element_data['ecp_electrons'] = n_elec

    # split the remaining lines by lines starting with a character
    ecp_potentials = helpers.partition_lines(element_lines[2:], lambda x: x[0].isalpha(), min_size=2)

    # Keep track of what the max AM we actually found is
    found_max = False
    for pot_lines in ecp_potentials:
        pot_am, pot_base_am = helpers.parse_line_regex(ecp_pot_am_re, pot_lines[0], 'ECP potential am')

        pot_am = lut.amchar_to_int(pot_am)

        if pot_base_am:
            pot_base_am = lut.amchar_to_int(pot_base_am[1:])  # Strip the - from the beginning

            if pot_base_am[0] != max_am:
                raise RuntimeError("Potential does not use max_am of {}. Uses {}".format(max_am, pot_base_am[0]))
        else:
            if found_max:
                raise RuntimeError("Found multiple potentials with single AM")

            if pot_am[0] != max_am:
                raise RuntimeError("Potential with single AM {} is not the same as lmax = {}".format(
                    pot_am[0], max_am))

            found_max = True

        ecp_data = helpers.parse_ecp_table(pot_lines[1:], order=['coeff', 'r_exp', 'g_exp'])
        ecp_pot = {
            'angular_momentum': pot_am,
            'ecp_type': 'scalar_ecp',
            'r_exponents': ecp_data['r_exp'],
            'gaussian_exponents': ecp_data['g_exp'],
            'coefficients': ecp_data['coeff']
        }

        element_data['ecp_potentials'].append(ecp_pot)


def _parse_ecp_lines(basis_lines, bs_data):
    # Strip all lines beginning with $
    basis_lines = helpers.prune_lines(basis_lines, '$')

    # Last line should be *
    # We don't need it
    if basis_lines[-1] != '*':
        raise RuntimeError("Missing terminating * line")
    basis_lines.pop()

    # Partition based on lines beginning with a character
    element_blocks = helpers.partition_lines(basis_lines, element_re.match, before=1)

    # Element lines should be surrounded by *
    # Check all first. the partition_lines above will eat part of the previous
    # element if the * is missing
    for element_lines in element_blocks:
        if element_lines[0] != '*':
            raise RuntimeError("Element line not preceded by *")
        if element_lines[2] != '*':
            raise RuntimeError("Element line not followed by *")

        # Check for any other lines starting with *
        for line in element_lines[3:]:
            if line.startswith('*'):
                raise RuntimeError("Found line starting with * that probably doesn't belong: " + line)

    # Now process all elements
    for element_lines in element_blocks:
        # Remove the two * lines and parse using the separate function
        element_lines = element_lines[1:2] + element_lines[3:]
        _parse_ecp_potential_lines(element_lines, bs_data)


def read_turbomole(basis_lines):
    '''Reads turbomole-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the turbomole format does not store all the fields we
       have, so some fields are left blank
    '''

    basis_lines = helpers.prune_lines(basis_lines, '#')

    # first line must begin with $, last line must be $end
    if basis_lines and not basis_lines[0][0].startswith('$'):
        raise RuntimeError("First line does not begin with $. Line: " + basis_lines[0])
    if basis_lines and basis_lines[-1] != '$end':
        raise RuntimeError("Last line of basis is not $end. Line: " + basis_lines[-1])

    bs_data = {}

    # Split into basis and ecp
    # Just split based on lines beginning with $
    basis_sections = helpers.partition_lines(basis_lines,
                                             lambda x: x.startswith('$') and x != '$end',
                                             min_blocks=1,
                                             max_blocks=2)

    for s in basis_sections:
        # Check if section is empty. If so, all lines start with $
        if all(x.startswith('$') for x in s):
            continue

        if len(s) == 0:  # Empty section
            continue
        elif s[0].lower() == '$ecp':
            _parse_ecp_lines(s, bs_data)
        elif section_re.match(s[0]):
            _parse_electron_lines(s, bs_data)
        else:
            raise RuntimeError("Unknown section " + s[0])

    return bs_data



"""
Tests functions for reading in formatted basis sets
"""

import basis_set_exchange as bse
import os
import pytest

from basis_set_exchange import convert
from .common_testvars import bs_read_formats, bs_write_formats, bs_names_sample

# We only want formats that we can actually write out and read from
my_write_formats = [x for x in bs_write_formats if x is not None]
my_read_formats = [x for x in bs_read_formats if x in my_write_formats]

# TODO - molcas inline vs. separate file
my_read_formats.remove('molcas')

@pytest.mark.parametrize('basis_name', bs_names_sample)
@pytest.mark.parametrize('fmt_from', my_read_formats)
@pytest.mark.parametrize('fmt_to', my_write_formats)
def test_convert_slow(basis_name, fmt_from, fmt_to, tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5

    # First, get the basis set in the format we want to read from
    bs_data = bse.get_basis(basis_name)
    in_file = bse.misc.basis_name_to_filename(basis_name)
    in_file += '.{}{}'.format(fmt_from, bse.writers.get_format_extension(fmt_from))
    in_file_path = os.path.join(tmp_path, in_file)
    bse.writers.write_formatted_basis_file(bs_data, in_file_path, fmt_from)

    # Now attempt the conversion
    out_file_path = in_file_path + '.{}{}'.format(fmt_to, bse.writers.get_format_extension(fmt_to))

    convert.convert_formatted_basis_file(in_file_path, out_file_path, fmt_from, fmt_to)



'''
Conversion of basis sets to cp2k format
'''

from .. import lut, sort, misc, printing


def write_cp2k(basis):
    '''Converts a basis set to cp2k format
    '''

    s = ''
    basis = sort.sort_basis(basis, True)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        for z in electron_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, normalize=True)
            elname = lut.element_name_from_Z(z, normalize=True)
            cont_string = misc.contraction_string(data)

            s += '# {} {} {}\n'.format(elname, basis['name'], cont_string)
            s += '{} {}\n'.format(sym, basis['name'])

            nshells = len(data['electron_shells'])
            s += '    {}\n'.format(nshells)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                am = shell['angular_momentum']
                min_am = min(am)
                max_am = max(am)
                ncont = len(coefficients)
                ncol = ncont + 1

                nprim = len(exponents)

                # First number is principle quantum number
                # But is not used, according to the documentation
                s += '{} {} {} {}'.format("1", min_am, max_am, nprim)

                if len(am) > 1:
                    for _ in am:
                        s += ' 1'
                else:
                    s += ' ' + str(ncont)
                s += '\n'

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places, convert_exp=False)
            s += '\n'

    # Write out ECP
    if ecp_elements:
        bsname = basis['name'].replace(' ', '_') + '_ECP'
        s += '\n\n## Effective core potentials\n'
        s += bsname + '\n'
        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, normalize=True)
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{} nelec {}\n'.format(sym, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).upper()

                if am[0] == max_ecp_am:
                    s += '{} ul\n'.format(sym)
                else:
                    s += '{} {}\n'.format(sym, amchar)

                point_places = [0, 9, 32]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places, convert_exp=False)

        s += 'END ' + bsname + '\n'

    return s



#!/usr/bin/env python3

# This script determines what basis sets are missing authoritative sources
# that should be placed in this directory

import os
from basis_set_exchange import get_metadata, misc

my_dir = os.path.dirname(os.path.abspath(__file__))
auth_sources = [x for x in os.listdir(my_dir) if x.endswith('.bz2')]

md = get_metadata()

found_ver = {}

# Start with everything in the metadata, removing what we find
# in this directory
missing = {k: list(v['versions'].keys()) for k, v in md.items()}

for s in auth_sources:
    bsname, ver, _, _ = s.split('.')

    if bsname not in missing:
        raise RuntimeError("Source {} does not correspond to a basis set in the library".format(bsname))
    if ver not in missing[bsname]:
        raise RuntimeError("Source {} version {} does not correspond to a basis set version in the library".format(
            bsname, ver))

    missing[bsname].remove(ver)

    # Also remove aliases
    for other_name in md[bsname]['other_names']:
        other_name = misc.transform_basis_name(other_name)
        missing[other_name].remove(ver)

missing = {k: v for k, v in missing.items() if v}

maxlen = max([len(k) for k in missing.keys()]) + 5

for k, v in missing.items():
    print('{:{maxlen}} {}'.format(k, ','.join(v), maxlen=maxlen))



"""
Tests for lookups of elemental data
"""

import pytest
from basis_set_exchange import lut, api
from .common_testvars import bs_metadata

ecp_basis_sets = [k for k,v in bs_metadata.items() if 'scalar_ecp' in v['function_types']]


@pytest.mark.slow
@pytest.mark.parametrize('basis_name', ecp_basis_sets)
def test_stored_nelec_start_slow(basis_name):
    bs_data = api.get_basis(basis_name)

    for el in bs_data['elements'].values():
        if 'ecp_electrons' not in el:
            continue

        ecp_electrons = el['ecp_electrons']
        starting_shells = lut.electron_shells_start(ecp_electrons, 8)

        # Make sure the number of covered electrons matches
        nelec_sum = 0
        for am, count in enumerate(starting_shells):
            # How many shells of AM are covered by the ECP
            covered = count - 1

            # Adjust for the principal quantum number where the shells for the AM start
            # (ie, p start at 2, d start at 3)
            covered -= am

            # Number of orbs = 2*am+1. Multiply by 2 to get electrons
            nelec_sum += (2 * am + 1) * 2 * covered

        assert nelec_sum == ecp_electrons




'''
Conversion of basis sets to Gaussian format
'''

from .. import lut, manip, sort, printing


def _write_g94_common(basis, add_harm_type, psi4_am):
    '''Converts a basis set to Gaussian format
    '''

    s = ''

    basis = manip.uncontract_general(basis, True)
    basis = manip.uncontract_spdf(basis, 1, False)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        for z in electron_elements:
            data = basis['elements'][z]

            sym = lut.element_sym_from_Z(z, True)
            s += '{}     0\n'.format(sym)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 1
                nprim = len(exponents)

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True).upper()

                if psi4_am and len(am) == 1 and am[0] >= 7:
                    # For am=7 and above, use explicit L={am} notation
                    amchar = "L=" + str(am[0])

                harm = ''
                if add_harm_type and shell['function_type'] == 'gto_cartesian':
                    harm = ' c'
                s += '{:4} {}   1.00{}\n'.format(amchar, nprim, harm)

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places, convert_exp=True)

            s += '****\n'

    # Write out ECP
    if ecp_elements:
        s += '\n'
        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])
            max_ecp_amchar = lut.amint_to_char([max_ecp_am], hij=True)

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{}     0\n'.format(sym)
            s += '{}-ECP     {}     {}\n'.format(sym, max_ecp_am, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']
                nprim = len(rexponents)

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True)

                if am[0] == max_ecp_am:
                    s += '{} potential\n'.format(amchar)
                else:
                    s += '{}-{} potential\n'.format(amchar, max_ecp_amchar)

                s += '  ' + str(nprim) + '\n'

                point_places = [0, 9, 32]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places, convert_exp=True)

    return s


def write_g94(basis):
    '''Converts a basis set to Gaussian format
    '''
    return _write_g94_common(basis, False, False)


def write_xtron(basis):
    '''Converts a basis set to xTron format

    xTron uses a modified gaussian format that puts 'c' on the same
    line as the angular momentum if the shell is cartesian.
    '''
    return _write_g94_common(basis, True, False)


def write_psi4(basis):
    '''Converts a basis set to Psi4 format

    Psi4 uses the same output as gaussian94, except
    that the first line must be cartesian/spherical,
    and it prefers to have a starting asterisks

    The cartesian/spherical line is added later, since it must
    be the first non-blank line.
    '''

    s = '****\n'
    s += _write_g94_common(basis, False, True)
    return s



"""
Tests BSE manip functions
"""

import os
import pytest

from basis_set_exchange import api, curate, manip, readers
from .common_testvars import bs_names_sample, dunningext_test_data_dir, truhlar_test_data_dir

dunningext_test_subdirs = os.listdir(dunningext_test_data_dir)
dunningext_test_subdirs = [x for x in dunningext_test_subdirs if x.startswith('aug-')]

truhlar_test_subdirs = os.listdir(truhlar_test_data_dir)
truhlar_test_subdirs = [x for x in truhlar_test_subdirs if x.startswith('cc-')]


@pytest.mark.parametrize('basis', bs_names_sample)
def test_manip_roundtrip(basis):
    bse_dict = api.get_basis(basis, uncontract_general=True, uncontract_spdf=True)
    bse_dict_gen = manip.make_general(bse_dict)
    bse_dict_unc = manip.uncontract_general(bse_dict_gen)

    assert curate.compare_basis(bse_dict, bse_dict_unc, rel_tol=0.0)


@pytest.mark.parametrize('testdir', dunningext_test_subdirs)
def test_manip_dunningext(testdir):
    full_testdir = os.path.join(dunningext_test_data_dir, testdir)
    basefile = testdir + '.nw.bz2'
    base_data = readers.read_formatted_basis_file(os.path.join(full_testdir, basefile))

    for level, prefix in [(2, 'd-'), (3, 't-'), (4, 'q-')]:
        ref = prefix + testdir + '.nw.ref.bz2'
        full_ref_path = os.path.join(full_testdir, ref)
        ref_data = readers.read_formatted_basis_file(full_ref_path, 'nwchem')
        ref_data = manip.make_general(ref_data)

        new_data = manip.extend_dunning_aug(base_data, level)
        new_data = manip.make_general(new_data)
        assert curate.compare_basis(new_data, ref_data)


@pytest.mark.parametrize('testdir', truhlar_test_subdirs)
def test_manip_truhlar(testdir):
    full_testdir = os.path.join(truhlar_test_data_dir, testdir)
    basefile = 'aug-' + testdir + '.nw.bz2'
    base_data = readers.read_formatted_basis_file(os.path.join(full_testdir, basefile))

    ref_files = os.listdir(full_testdir)
    ref_files = [x for x in ref_files if x.endswith('.ref.bz2')]

    # Made a mistake here once where reference files weren't being read
    assert ref_files

    for ref in ref_files:
        month = ref.split('-')[0]
        full_ref_path = os.path.join(full_testdir, ref)
        ref_data = readers.read_formatted_basis_file(full_ref_path, 'nwchem')
        ref_data = manip.make_general(ref_data)

        new_data = manip.truhlar_calendarize(base_data, month, use_copy=True)
        assert curate.compare_basis(new_data, ref_data)



'''
Miscellaneous curation functions
'''

from .. import fileio
from .. import misc


def elements_in_files(filelist):
    '''Get a list of what elements exist in JSON files

    This works on table, element, and component data files

    Parameters
    ----------
    filelist : list
        A list of paths to json files

    Returns
    -------
    dict
        Keys are the file path, value is a compacted element string of
        what elements are in that file
    '''

    ret = {}
    for fpath in filelist:
        filedata = fileio.read_json_basis(fpath)
        els = list(filedata['elements'].keys())
        ret[fpath] = misc.compact_elements(els)

    return ret


def component_file_refs(filelist):
    '''Get a list of what elements/refrences exist in component JSON files

    Parameters
    ----------
    filelist : list
        A list of paths to json files

    Returns
    -------
    dict
        Keys are the file path, value is a list of tuples (compacted element string, refs tuple)
    '''

    ret = {}
    for fpath in filelist:
        filedata = fileio.read_json_basis(fpath)

        refdict = {}
        for el, eldata in filedata['elements'].items():
            refs = tuple(eldata['references'])
            if refs not in refdict:
                refdict[refs] = [el]
            else:
                refdict[refs].append(el)

        entry = []
        for k, v in refdict.items():
            entry.append((misc.compact_elements(v), k))

        ret[fpath] = entry

    return ret



'''
Conversion of basis sets to Molpro format
'''

from .. import lut, manip, misc, sort
from .common import find_range


def write_molpro(basis):
    '''Converts a basis set to Molpro format
    '''

    # Uncontract all, and make as generally-contracted as possible
    basis = manip.make_general(basis, False, True)
    basis = sort.sort_basis(basis, True)

    s = ''

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    if electron_elements:
        # basis set starts with a string
        s += 'basis={\n'

        # Electron Basis
        for z in electron_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            s += '!\n'
            s += '! {:20} {}\n'.format(lut.element_name_from_Z(z), misc.contraction_string(data))

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am).lower()
                s += '{}, {} , {}\n'.format(amchar, sym, ', '.join(exponents))
                for c in coefficients:
                    first, last = find_range(c)
                    s += 'c, {}.{}, {}\n'.format(first + 1, last + 1, ', '.join(c[first:last + 1]))
        s += '}\n'

    # Write out ECP
    if ecp_elements:
        s += '\n\n! Effective core Potentials\n'

        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).lower()
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += 'ECP, {}, {}, {} ;\n'.format(sym, data['ecp_electrons'], max_ecp_am)

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).lower()
                s += '{};'.format(len(rexponents))

                if am[0] == max_ecp_am:
                    s += ' !  ul potential\n'
                else:
                    s += ' !  {}-ul potential\n'.format(amchar)

                for p in range(len(rexponents)):
                    s += '{},{},{};\n'.format(rexponents[p], gexponents[p], coefficients[0][p])
    return s



'''
Conversion of references to bibtex format
'''


def write_bib(key, ref):
    '''Convert a single reference to bibtex format
    '''
    s = ''

    s += '@{}{{{},\n'.format(ref['_entry_type'], key)

    entry_lines = []
    for k, v in ref.items():
        if k == '_entry_type':
            continue

        # Handle authors/editors
        if k == 'authors':
            entry_lines.append('    author = {{{}}}'.format(' and '.join(v)))
        elif k == 'editors':
            entry_lines.append('    editor = {{{}}}'.format(' and '.join(v)))
        else:
            entry_lines.append('    {} = {{{}}}'.format(k, v))

    s += ',\n'.join(entry_lines)
    s += '\n}'

    return s



'''
Conversion of basis sets to cfour/aces2/genbas format
'''

import math
from .. import lut, manip, sort, printing


def _cfour_exp(e):
    '''Formats an exponent for CFour'''
    return e.replace('E', 'D') + ' '


def _cfour_coef(c):
    '''Formats a coefficient for CFour'''
    return c.replace('E', 'D') + ' '


def _aces_exp(e):
    '''Formats an exponent for AcesII'''

    e = float(e)
    # Some basis sets have negative exponents???
    mag = int(math.log(abs(e), 10))

    mag = max(mag, 1)

    # Make room for the negative sign
    if e < 0.0:
        mag += 1

    # Number of decimal places to show
    ndec = min(7, 14 - 2 - mag)

    fmtstr = '{{:14.{}f}}'.format(ndec)
    s = fmtstr.format(e)

    # Trim a single trailing zero if there is one
    # and our string takes up all 14 characters
    if s[0] != ' ' and s[-1] == '0':
        s = ' ' + s[:-1]

    return s


def _aces_coef(c):
    '''Formats a coefficient for AcesII'''
    c = float(c)
    return '{:10.7f} '.format(c)


def _print_columns(data, ncol):
    s = ''
    for i in range(0, len(data), ncol):
        s += ''.join(data[i:i+ncol]) + '\n'
    return s


def _write_genbas_internal(basis, exp_formatter, coef_formatter):
    # Uncontract all, then make general
    basis = manip.make_general(basis, False, True)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    s = '\n'

    if electron_elements:
        # Electron Basis
        for z in electron_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            nshell = len(data['electron_shells'])

            s += '{}:{}\n'.format(sym, basis['name'])
            s += basis['description'] + '\n'
            s += '\n'
            s += '{:>3}\n'.format(nshell)

            s_am = ''
            s_ngen = ''
            s_nprim = ''
            for sh in data['electron_shells']:
                s_am += '{:>5}'.format(sh['angular_momentum'][0])
                s_ngen += '{:>5}'.format(len(sh['coefficients']))
                s_nprim += '{:>5}'.format(len(sh['exponents']))

            s += s_am + '\n'
            s += s_ngen + '\n'
            s += s_nprim + '\n'
            s += '\n'

            for shell in data['electron_shells']:
                exponents = [exp_formatter(x) for x in shell['exponents']]
                coefficients = [[coef_formatter(x) for x in y] for y in shell['coefficients']]
                coefficients = list(map(list, zip(*coefficients)))

                s += _print_columns(exponents, 5) + '\n'
                for c in coefficients:
                    s += _print_columns(c, 7)
                s += '\n'

    # Write out ECP
    if ecp_elements:
        s += '\n\n! Effective core Potentials\n'

        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])
            max_ecp_amchar = lut.amint_to_char([max_ecp_am]).lower()

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '*\n'
            s += '{}:{}\n'.format(sym, basis['name'])
            s += '# ' + basis['description'] + '\n'
            s += '*\n'
            s += '    NCORE = {}    LMAX = {}\n'.format(data['ecp_electrons'], max_ecp_am)

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).lower()

                if am[0] == max_ecp_am:
                    s += '{}\n'.format(amchar)
                else:
                    s += '{}-{}\n'.format(amchar, max_ecp_amchar)

                point_places = [6, 18, 25]
                s += printing.write_matrix([*coefficients, rexponents, gexponents], point_places)
                #for p in range(len(rexponents)):
                #    s += '{}  {}  {};\n'.format(gexponents[p], rexponents[p], coefficients[0][p])
            s += '*\n'
    return s


def write_cfour(basis):
    '''Converts a basis set to cfour
    '''

    # March 2019
    # Format determined from http://slater.chemie.uni-mainz.de/cfour/index.php?n=Main.NewFormatOfAnEntryInTheGENBASFile

    return _write_genbas_internal(basis, _cfour_exp, _cfour_coef)


def write_aces2(basis):
    '''Converts a basis set to cfour
    '''

    # March 2019
    # Format determined from http://slater.chemie.uni-mainz.de/cfour/index.php?n=Main.OldFormatOfAnEntryInTheGENBASFile

    return _write_genbas_internal(basis, _aces_exp, _aces_coef)



'''
Add a basis set to the library
'''

import os
import datetime
from ..fileio import read_json_basis, write_json_basis
from ..misc import expand_elements
from ..validator import validate_data
from ..skel import create_skel
from ..readers import read_formatted_basis_file
from .metadata import create_metadata_file


def add_from_components(component_files, data_dir, subdir, file_base, name, family, role, description, version,
                        revision_description):
    '''
    Add a basis set to this library that is a combination of component files

    This takes in a list of component basis files and creates a new basis set for the intersection
    of all the elements contained in those files.  This creates the element, and table basis set
    files in the given data_dir (and subdir). The metadata file for the basis is created if it
    doesn't exist, and the main metadata file is also updated.

    Parameters
    ----------
    component_files : str
        Path to component json files (in BSE format already)
    data_dir : str
        Path to the data directory to add the data to
    subdir : str
        Subdirectory of the data directory to add the basis set to
    file_base : str
        Base name for new files
    name : str
        Name of the basis set
    family : str
        Family to which this basis set belongs
    role : str
        Role of the basis set (orbital, etc)
    description : str
        Description of the basis set
    version : str
        Version of the basis set
    revision_description : str
        Description of this version of the basis set
    '''

    if not component_files:
        raise RuntimeError("Need at least one component file to create a basis set from")

    # Determine what files have which elements
    valid_elements = None

    # And the relative path of the component files to the data dir
    cfile_relpaths = []

    for cfile in component_files:
        cfile_data = read_json_basis(cfile)
        cfile_elements = set(cfile_data['elements'].keys())

        relpath = os.path.relpath(cfile, data_dir)

        if valid_elements is None:
            valid_elements = cfile_elements
        else:
            valid_elements = valid_elements.intersection(cfile_elements)

        cfile_relpaths.append(relpath)

    valid_elements = list(sorted(valid_elements, key=lambda x: int(x)))

    # Start the data files for the element and table json
    element_file_data = create_skel('element')
    element_file_data['name'] = name
    element_file_data['description'] = description
    element_file_name = '{}.{}.element.json'.format(file_base, version)
    element_file_relpath = os.path.join(subdir, element_file_name)
    element_file_path = os.path.join(data_dir, element_file_relpath)

    table_file_data = create_skel('table')
    table_file_data['revision_description'] = revision_description
    table_file_data['revision_date'] = datetime.date.today().isoformat()
    table_file_name = '{}.{}.table.json'.format(file_base, version)

    # and the metadata file
    meta_file_data = create_skel('metadata')
    meta_file_data['names'] = [name]
    meta_file_data['family'] = family
    meta_file_data['description'] = description
    meta_file_data['role'] = role
    meta_file_name = '{}.metadata.json'.format(file_base)

    # These get created directly in the top-level data directory
    table_file_path = os.path.join(data_dir, table_file_name)
    meta_file_path = os.path.join(data_dir, meta_file_name)

    # Can just make all the entries for the table file pretty easily
    # (we add the relative path to the location of the element file,
    # which resides in subdir)
    table_file_entry = element_file_relpath
    table_file_data['elements'] = {k: table_file_entry for k in valid_elements}

    # Add to the element file data
    for el in valid_elements:
        element_file_data['elements'][el] = {'components': cfile_relpaths}

    # Verify all data using the schema
    validate_data('element', element_file_data)
    validate_data('table', table_file_data)

    ######################################################################################
    # Before creating any files, check that all the files don't already exist.
    # Yes, there is technically a race condition (files could be created between the
    # check and then actually writing out), but if that happens, you are probably using
    # this library wrong
    #
    # Note that the metadata file may exist already. That is ok
    ######################################################################################
    if os.path.exists(element_file_path):
        raise RuntimeError("Element json file {} already exists".format(element_file_path))
    if os.path.exists(table_file_path):
        raise RuntimeError("Table json file {} already exists".format(table_file_path))

    #############################################
    # Actually create all the files
    #############################################

    # First, create the subdirectory
    subdir_path = os.path.join(data_dir, subdir)
    if not os.path.exists(subdir_path):
        os.makedirs(subdir_path)

    write_json_basis(element_file_path, element_file_data)
    write_json_basis(table_file_path, table_file_data)

    # Create the metadata file if it doesn't exist already
    if not os.path.exists(meta_file_path):
        write_json_basis(meta_file_path, meta_file_data)

    # Update the metadata file
    metadata_file = os.path.join(data_dir, 'METADATA.json')
    create_metadata_file(metadata_file, data_dir)


def add_basis(bs_file,
              data_dir,
              subdir,
              file_base,
              name,
              family,
              role,
              description,
              version,
              revision_description,
              data_source,
              refs=None,
              file_fmt=None):
    '''
    Add a basis set to this library

    This takes in a single file containing the basis set is some format, parses it, and
    create the component, element, and table basis set files in the given data_dir (and subdir).
    The metadata file for the basis is created if it doesn't exist, and the main metadata file is
    also updated.

    Parameters
    ----------
    bs_file : str
        Path to the file with formatted basis set information
    data_dir : str
        Path to the data directory to add the data to
    subdir : str
        Subdirectory of the data directory to add the basis set to
    file_base : str
        Base name for new files
    name : str
        Name of the basis set
    family : str
        Family to which this basis set belongs
    role : str
        Role of the basis set (orbital, etc)
    description : str
        Description of the basis set
    version : str
        Version of the basis set
    revision_description : str
        Description of this version of the basis set
    data_source : str
        Description of where this data came from
    refs : dict or str
        Mapping of references to elements. This can be a dictionary with a compressed
        string of elements as keys and a list of reference strings as values.
        For example, {'H,Li-B,Kr': ['kumar2018a']}

        If a list or string is passed, then those reference(s) will be used for
        all elements.

        Elements that exist in the file but do not have a reference are given the
        usual 'noref' extension and the references entry is empty.
    file_fmt : str
        Format of the input basis data (None = autodetect)
    '''

    # Read the basis set data into a component file, and add the description
    bs_data = read_formatted_basis_file(bs_file, file_fmt, validate=True, as_component=True)
    bs_data['description'] = description
    bs_data['data_source'] = data_source

    if refs is None:
        refs = []

    # We keep track of which elements we've done so that
    # we can detect duplicates in the references (which would be an error)
    # (and also handle elements with no reference)
    orig_elements = bs_data['elements']
    done_elements = []

    # If a string or list of strings, use that as a reference for all elements
    if isinstance(refs, str):
        for k, v in bs_data['elements'].items():
            v['references'] = [refs]
    elif isinstance(refs, list):
        for k, v in bs_data['elements'].items():
            v['references'] = refs
    else:
        for k, v in refs.items():
            # Expand the string a list of integers (as strings)
            elements = expand_elements(k, True)

            # Make sure we have info for the given elements
            # and that there are no duplicates
            for el in elements:
                if el not in orig_elements:
                    raise RuntimeError("Element {} not found in file {}".format(el, bs_file))
                if el in done_elements:
                    raise RuntimeError("Duplicate element {} in reference string {}".format(el, k))

                if isinstance(v, str):
                    bs_data['elements'][el]['references'] = [v]
                else:
                    bs_data['elements'][el]['references'] = v

            done_elements.extend(elements)

        # Handle elements without a reference
        noref_elements = set(orig_elements.keys()) - set(done_elements)

        if noref_elements:
            for el in noref_elements:
                bs_data['elements'][el]['references'] = []

    # Create the filenames for the components
    # Also keep track of where data for each element is (for the element and table files)
    component_file_name = file_base + '.' + str(version) + '.json'
    component_file_relpath = os.path.join(subdir, component_file_name)
    component_file_path = os.path.join(data_dir, component_file_relpath)

    # Verify all data using the schema
    validate_data('component', bs_data)

    ######################################################################################
    # Before creating any files, check that all the files don't already exist.
    # Yes, there is technically a race condition (files could be created between the
    # check and then actually writing out), but if that happens, you are probably using
    # this library wrong
    #
    # Note that the metadata file may exist already. That is ok
    ######################################################################################
    if os.path.exists(component_file_path):
        raise RuntimeError("Component json file {} already exists".format(component_file_path))

    #############################################
    # Actually create all the files
    #############################################

    # First, create the subdirectory
    subdir_path = os.path.join(data_dir, subdir)
    if not os.path.exists(subdir_path):
        os.makedirs(subdir_path)

    write_json_basis(component_file_path, bs_data)

    # Do all the rest
    add_from_components([component_file_path], data_dir, subdir, file_base, name, family, role, description, version,
                        revision_description)



'''
Conversion of basis sets to the BSE debug/dump format
'''

from ..printing import element_data_str


def write_bsedebug(basis):
    '''Converts a basis set to BSE Debug format
    '''

    s = ''

    for el, eldata in basis['elements'].items():
        s += element_data_str(el, eldata)
    return s



'''
Basis Set Exchange Python library

This package contains utilities for obtaining, reading, writing,
and converting basis set information
'''

# Check to make sure we are using python 3
import sys

if sys.version_info < (3, 0):
    raise RuntimeError("This library requires python 3")

# Just import the basic user API
from .api import (get_basis, lookup_basis_by_role, get_metadata, get_reference_data, get_all_basis_names,
                  get_references, get_basis_family, filter_basis_sets, get_families, get_family_notes, get_basis_notes,
                  has_basis_notes, has_family_notes, get_roles, get_formats,
                  version, get_data_dir)

from .readers import read_formatted_basis_file, read_formatted_basis_str, get_reader_formats
from .writers import write_formatted_basis_file, write_formatted_basis_str, get_writer_formats
from .convert import convert_formatted_basis_file, convert_formatted_basis_str
from .refconverters import get_reference_formats
from .validator import validate_file, validate_data

from .bundle import create_bundle, get_archive_types

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



'''
Conversion of basis sets to Molcas format
'''

from .. import lut, manip, printing, misc, sort, misc


def write_molcas(basis):
    '''Converts a basis set to Molcas format
    '''

    basis = manip.make_general(basis, False, True)
    basis = sort.sort_basis(basis, False)

    s = ''

    for z, data in basis['elements'].items():
        s += 'Basis set\n'
        has_electron = 'electron_shells' in data
        has_ecp = 'ecp_potentials' in data

        el_name = lut.element_name_from_Z(z).upper()
        el_sym = lut.element_sym_from_Z(z, normalize=True)
        s += '* {}  {}\n'.format(el_name, misc.contraction_string(data))

        # if ECP is present, the line should be "{sym}.ECP /inline"
        ecp_tag = '.ECP' if has_ecp else ''
        s += ' {}{}    / inline\n'.format(el_sym, ecp_tag)

        if has_electron:
            max_am = misc.max_am(data['electron_shells'])

            # number of electrons
            # should be z - number of ecp electrons
            nelectrons = int(z)
            if has_ecp:
                nelectrons -= data['ecp_electrons']

            s += '{:>7}.00   {}\n'.format(nelectrons, max_am)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                nprim = len(exponents)
                ngen = len(coefficients)

                amchar = lut.amint_to_char(shell['angular_momentum']).upper()
                s += '* {}-type functions\n'.format(amchar)
                s += '{:>6}    {}\n'.format(nprim, ngen)

                s += printing.write_matrix([exponents], [17])

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ngen + 1)]
                s += printing.write_matrix(coefficients, point_places)

        if has_ecp:
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += 'PP, {}, {}, {} ;\n'.format(el_sym, data['ecp_electrons'], max_ecp_am)

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am)
                s += '{};'.format(len(rexponents))

                if am[0] == max_ecp_am:
                    s += ' !  ul potential\n'
                else:
                    s += ' !  {}-ul potential\n'.format(amchar)

                for p in range(len(rexponents)):
                    s += '{},{},{};\n'.format(rexponents[p], gexponents[p], coefficients[0][p])

            s += 'Spectral\n'
            s += 'End of Spectral\n'
            s += '*\n'

        s += 'End of basis set\n\n'

    return s



"""
Common basis set manipulations

This module contains functions for uncontracting and merging basis set
data, as well as some other small functions.
"""

import copy
from . import lut, skel, misc


def merge_element_data(dest, sources, use_copy=True):
    """
    Merges the basis set data for an element from multiple sources
    into dest.

    The destination is not modified, and a (shallow) copy of dest is returned
    with the data from sources added.

    If use_copy is True, then the data merged into dest will be a (deep)
    copy of that found in sources. Otherwise, data may be shared between dest
    and sources
    """

    if dest is not None:
        ret = dest.copy()
    else:
        ret = {}

    if use_copy:
        sources = copy.deepcopy(sources)

    # Note that we are not copying notes/data_sources
    for s in sources:
        if 'electron_shells' in s:
            if 'electron_shells' not in ret:
                ret['electron_shells'] = []
            ret['electron_shells'].extend(s['electron_shells'])
        if 'ecp_potentials' in s:
            if 'ecp_potentials' in ret:
                raise RuntimeError('Cannot overwrite existing ECP')
            ret['ecp_potentials'] = s['ecp_potentials']
            ret['ecp_electrons'] = s['ecp_electrons']
        if 'references' in s:
            if 'references' not in ret:
                ret['references'] = []
            for ref in s['references']:
                ret['references'].append(ref)

    return ret


def prune_shell(shell, use_copy=True):
    """
    Removes exact duplicates of primitives, and condenses duplicate exponents
    into general contractions

    Also removes primitives if all coefficients are zero
    """

    new_exponents = []
    new_coefficients = []

    exponents = shell['exponents']
    nprim = len(exponents)

    # transpose of the coefficient matrix
    coeff_t = list(map(list, zip(*shell['coefficients'])))

    # Group by exponents
    ex_groups = []
    for i in range(nprim):
        for ex in ex_groups:
            if float(exponents[i]) == float(ex[0]):
                ex[1].append(coeff_t[i])
                break
        else:
            ex_groups.append((exponents[i], [coeff_t[i]]))

    # Now collapse within groups
    for ex in ex_groups:
        if len(ex[1]) == 1:
            # only add if there is a nonzero contraction coefficient
            if not all([float(x) == 0.0 for x in ex[1][0]]):
                new_exponents.append(ex[0])
                new_coefficients.append(ex[1][0])
            continue

        # ex[1] contains rows of coefficients. The length of ex[1]
        # is the number of times the exponent is duplicated. Columns represent general contractions.
        # We want to find the non-zero coefficient in each column, if it exists
        # The result is a single row with a length representing the number
        # of general contractions

        new_coeff_row = []

        # so take yet another transpose.
        ex_coeff = list(map(list, zip(*ex[1])))
        for g in ex_coeff:
            nonzero = [x for x in g if float(x) != 0.0]
            if len(nonzero) > 1:
                raise RuntimeError("Exponent {} is duplicated within a contraction".format(ex[0]))

            if not nonzero:
                new_coeff_row.append(g[0])
            else:
                new_coeff_row.append(nonzero[0])

        # only add if there is a nonzero contraction coefficient anywhere for this exponent
        if not all([float(x) == 0.0 for x in new_coeff_row]):
            new_exponents.append(ex[0])
            new_coefficients.append(new_coeff_row)

    # take the transpose again, putting the general contraction
    # as the slowest index
    new_coefficients = list(map(list, zip(*new_coefficients)))

    shell['exponents'] = new_exponents
    shell['coefficients'] = new_coefficients

    return shell


def prune_basis(basis, use_copy=True):
    """
    Removes primitives that have a zero coefficient, and
    removes duplicate primitives and shells

    This only finds EXACT duplicates, and is meant to be used
    after other manipulations

    If use_copy is True, the input basis set is not modified.
    """

    if use_copy:
        basis = copy.deepcopy(basis)

    for k, el in basis['elements'].items():
        if 'electron_shells' not in el:
            continue

        shells = el.pop('electron_shells')
        shells = [prune_shell(sh, False) for sh in shells]

        # Remove any duplicates
        el['electron_shells'] = []

        for sh in shells:
            if sh not in el['electron_shells']:
                el['electron_shells'].append(sh)

    return basis


def uncontract_spdf(basis, max_am=0, use_copy=True):
    """
    Removes sp, spd, spdf, etc, contractions from a basis set

    The general contractions are replaced by uncontracted versions

    Contractions up to max_am will be left in place. For example,
    if max_am = 1, spd will be split into sp and d

    The input basis set is not modified. The returned basis
    may have functions with coefficients of zero and may have duplicate
    shells.

    If use_copy is True, the input basis set is not modified.
    """

    if use_copy:
        basis = copy.deepcopy(basis)

    for k, el in basis['elements'].items():

        if 'electron_shells' not in el:
            continue
        newshells = []

        for sh in el['electron_shells']:

            # am will be a list
            am = sh['angular_momentum']
            coeff = sh['coefficients']

            # if this is an sp, spd,...  orbital
            if len(am) > 1:
                newsh = sh.copy()
                newsh['angular_momentum'] = []
                newsh['coefficients'] = []

                ngen = len(sh['coefficients'])
                for g in range(ngen):
                    if am[g] > max_am:
                        newsh2 = sh.copy()
                        newsh2['angular_momentum'] = [am[g]]
                        newsh2['coefficients'] = [coeff[g]]
                        newshells.append(newsh2)
                    else:
                        newsh['angular_momentum'].append(am[g])
                        newsh['coefficients'].append(coeff[g])

                newshells.insert(0, newsh)

            else:
                newshells.append(sh)

        el['electron_shells'] = newshells

    return basis


def uncontract_general(basis, use_copy=True):
    """
    Removes the general contractions from a basis set

    The input basis set is not modified. The returned basis
    may have functions with coefficients of zero and may have duplicate
    shells.

    If use_copy is True, the input basis set is not modified.
    """

    if use_copy:
        basis = copy.deepcopy(basis)

    for k, el in basis['elements'].items():

        if 'electron_shells' not in el:
            continue

        newshells = []

        for sh in el['electron_shells']:
            # See if we actually have to uncontract
            # Also, don't uncontract sp, spd,.... orbitals
            #      (leave that to uncontract_spdf)
            if len(sh['coefficients']) == 1 or len(sh['angular_momentum']) > 1:
                newshells.append(sh)
            else:
                if len(sh['angular_momentum']) == 1:
                    for c in sh['coefficients']:
                        # copy, them replace 'coefficients'
                        newsh = sh.copy()
                        newsh['coefficients'] = [c]
                        newshells.append(newsh)

        el['electron_shells'] = newshells

    # If use_basis is True, we already made our deep copy
    return prune_basis(basis, False)


def uncontract_segmented(basis, use_copy=True):
    """
    Removes the segmented contractions from a basis set

    This implicitly removes general contractions as well,
    but will leave sp, spd, ... orbitals alone

    The input basis set is not modified. The returned basis
    may have functions with coefficients of zero and may have duplicate
    shells.

    If use_copy is True, the input basis set is not modified.
    """

    if use_copy:
        basis = copy.deepcopy(basis)

    for k, el in basis['elements'].items():

        if 'electron_shells' not in el:
            continue

        newshells = []

        for sh in el['electron_shells']:
            exponents = sh['exponents']
            nam = len(sh['angular_momentum'])

            for i in range(len(exponents)):
                newsh = sh.copy()
                newsh['exponents'] = [exponents[i]]
                newsh['coefficients'] = [["1.00000000E+00"] * nam]

                # Remember to transpose the coefficients
                newsh['coefficients'] = list(map(list, zip(*newsh['coefficients'])))

                newshells.append(newsh)

        el['electron_shells'] = newshells

    return basis


def make_general(basis, skip_spdf=False, use_copy=True):
    """
    Makes one large general contraction for each angular momentum

    If use_copy is True, the input basis set is not modified.

    The output of this function is not pretty. If you want to make it nicer,
    use sort_basis afterwards.
    """

    zero = '0.00000000'

    if use_copy:
        basis = copy.deepcopy(basis)

    if not skip_spdf:
        basis = uncontract_spdf(basis, 0, False)

    for k, el in basis['elements'].items():
        if 'electron_shells' not in el:
            continue

        newshells = []

        # See what we have
        all_am = []
        for sh in el['electron_shells']:
            am = sh['angular_momentum']

            # Skip sp shells
            if len(am) > 1:
                newshells.append(sh)
                continue

            if am not in all_am:
                all_am.append(am)

        all_am = sorted(all_am)

        for am in all_am:
            newsh = {
                'angular_momentum': am,
                'exponents': [],
                'coefficients': [],
                'region': '',
                'function_type': None,
            }

            # Do exponents first
            for sh in el['electron_shells']:
                if sh['angular_momentum'] == am:
                    newsh['exponents'].extend(sh['exponents'])

            # Number of primitives in the new shell
            nprim = len(newsh['exponents'])

            cur_prim = 0
            for sh in el['electron_shells']:
                if sh['angular_momentum'] != am:
                    continue

                if newsh['function_type'] is None:
                    newsh['function_type'] = sh['function_type']

                # Make sure the shells we are merging have the same function types
                ft1 = newsh['function_type']
                ft2 = sh['function_type']

                # Check if one function type is the subset of another
                # (should handle gto/gto_spherical, etc)
                if ft1 not in ft2 and ft2 not in ft1:
                    raise RuntimeError("Cannot make general contraction of different function types")

                ngen = len(sh['coefficients'])

                for g in range(ngen):
                    coef = [zero] * cur_prim
                    coef.extend(sh['coefficients'][g])
                    coef.extend([zero] * (nprim - len(coef)))
                    newsh['coefficients'].append(coef)

                cur_prim += len(sh['exponents'])

            newshells.append(newsh)

        el['electron_shells'] = newshells

    return basis


def _is_single_column(col):
    return sum(float(x) != 0.0 for x in col) == 1


def optimize_general(basis, use_copy=True):
    """
    Optimizes the general contraction using the method of Hashimoto et al

    .. seealso :: | T. Hashimoto, K. Hirao, H. Tatewaki
                  | 'Comment on Dunning's correlation-consistent basis set'
                  | Chemical Physics Letters v243, Issues 1-2, pp, 190-192 (1995)
                  | https://doi.org/10.1016/0009-2614(95)00807-G

    """

    if use_copy:
        basis = copy.deepcopy(basis)

    # Make as generally-contracted as possible first
    basis = make_general(basis, skip_spdf=True, use_copy=False)

    for eldata in basis['elements'].values():

        if not 'electron_shells' in eldata:
            continue

        elshells = eldata['electron_shells']
        for sh in elshells:
            exponents = sh['exponents']
            coefficients = sh['coefficients']
            nprim = len(exponents)
            nam = len(sh['angular_momentum'])

            # Skip sp shells and shells with only one general contraction
            if nam > 1 or len(coefficients) < 2:
                continue

            # First, find columns (general contractions) with a single non-zero value
            single_columns = [idx for idx, c in enumerate(coefficients) if _is_single_column(c)]

            # Find the corresponding rows that have a value in one of these columns
            # Note that at this stage, the row may have coefficients in more than one
            # column. That is what we are looking for

            # Also, test to see that each row is only represented once. That is, there should be
            # no rows that are part of single columns (this would represent duplicate shells).
            # This can happen in poorly-formatted basis sets and is an error
            row_col_pairs = []
            all_row_idx = []
            for col_idx in single_columns:
                col = coefficients[col_idx]
                for row_idx in range(nprim):
                    if float(col[row_idx]) != 0.0:
                        if row_idx in all_row_idx:
                            raise RuntimeError("Badly-formatted basis. Row {} makes duplicate shells".format(row_idx))

                        # Store the index of the nonzero value in single_columns
                        row_col_pairs.append((row_idx, col_idx))
                        all_row_idx.append(row_idx)

            # Now for each row/col pair, zero out the entire row
            # EXCEPT for the column that has the single value
            for row_idx, col_idx in row_col_pairs:
                for idx, col in enumerate(coefficients):
                    if float(col[row_idx]) != 0.0 and col_idx != idx:
                        col[row_idx] = '0.0000000E+00'

    return basis


def extend_dunning_aug(basis, level, use_copy=True, as_component=False):
    '''
    Extends the augmenting functions of a dunning basis set (aug -> daug,taug,...)

    Parameters
    ----------
    basis : dict
        Basis set dictionary to work with
    level: int
        Level to create (must be >1). 2 = daug, 3 = taug, etc
    use_copy: bool
        If True, the input basis set is not modified.
    '''

    if level < 2:
        raise RuntimeError("Level = {} is invalid for dunning_extend_aug".format(level))

    # We need to combine shells by AM
    # I guess we don't NEED to, but it makes things a lot easier
    basis_copy = make_general(basis, use_copy=True)

    # We will now store the new shells in 'basis'
    # If as_component is specified, then create a new empty component basis
    if as_component:
        basis = skel.create_skel('component')
        basis['elements'] = {k: {} for k, v in basis_copy['elements'].items() if 'electron_shells' in v}
    elif use_copy:
        basis = copy.deepcopy(basis)

    # From Woon & Dunning, Jr
    # J. Chem. Phys. v100, No. 4, p. 2975 (1994)
    # DOI: 10.1063/1.466439
    #
    # The exponent for d-aug-cc-pVXZ: alpha*beta
    #                  t-aug-cc-pVXZ: alpha*(beta**2)
    # and so on.
    #
    # alpha = smallest exponent in aug-cc-pVXZ
    # beta = ratio of two most diffuse functions (beta < 1)
    #
    # This applies to all angular momentum (which have been combined into
    # shells already)

    for el_z, eldata in basis_copy['elements'].items():
        if 'electron_shells' not in eldata:
            continue

        el_sym = lut.element_sym_from_Z(el_z)

        new_shells = []

        for shell in eldata['electron_shells']:
            # Find the two smallest exponents. The smallest is alpha
            # beta is the ratio alpha/(next smallest)
            # Keep track of the indices as well
            exponents = [(float(x), idx) for idx, x in enumerate(shell['exponents'])]
            sorted_exponents = sorted(exponents)

            if len(sorted_exponents) < 2:
                raise RuntimeError(
                    "Need more than two exponents to extend dunning augmentation. Element {} has {}: ".format(
                        el_sym, ','.join(shell['exponents'])))

            alpha, alpha_idx = sorted_exponents[0]
            beta_tmp, _ = sorted_exponents[1]
            beta = alpha / beta_tmp

            if alpha == beta_tmp:
                raise RuntimeError(
                    "Two smallest exponents are the same. Duplicate exponents are not a good thing here. Exponent: {}".
                    format(alpha))

            # Test that the primitive for alpha is completely uncontracted
            alpha_coefs = [float(c[alpha_idx]) for c in shell['coefficients']]
            n_coefs = len(alpha_coefs)
            if alpha_coefs.count(0.0) != n_coefs - 1 or alpha_coefs.count(1.0) != 1:
                raise RuntimeError("Smallest exponent is not completely uncontracted. Exponent: {}".format(alpha))

            new_exponents = []
            for i in range(1, level):
                new_exponents.append(alpha * (beta**i))

            new_exponents = ['{:.3e}'.format(x) for x in new_exponents]

            # add the new exponents as new uncontracted shells
            for ex in new_exponents:
                new_shells.append({
                    'function_type': shell['function_type'],
                    'region': shell['region'],
                    'angular_momentum': shell['angular_momentum'],
                    'exponents': [ex],
                    'coefficients': [['1.00000000']]
                })

        # add the shells to the original basis set
        # (if this is a component basis, then 'electron_shells' may not exist)
        if 'electron_shells' not in basis['elements'][el_z]:
            basis['elements'][el_z]['electron_shells'] = []

        basis['elements'][el_z]['electron_shells'].extend(new_shells)

    return basis


def remove_primitive(electron_shell, idx_to_remove):
    '''Removes a primitive (by index) of a electron_shell

    The electron_shell passed in is modified
    '''

    electron_shell['exponents'].pop(idx_to_remove)
    for c in electron_shell['coefficients']:
        c.pop(idx_to_remove)

    # We may have removed the only primitive of a general contraction
    # So remove general contractions if all coefficients are zero
    electron_shell['coefficients'] = [
        gen for gen in electron_shell['coefficients'] if any([float(c) != 0.0 for c in gen])
    ]


def _element_remove_diffuse(eldata, nremove):
    if 'electron_shells' not in eldata:
        pass

    shells = eldata['electron_shells']
    max_am = misc.max_am(shells)
    if nremove == 'all':
        nremove = max_am + 1

    am_toremove = list(range(max_am, max_am - nremove, -1))

    # We may be asked to remove more than we have. That is ok
    am_toremove = [x for x in am_toremove if x >= 0]

    for shell in shells:
        shell_am = shell['angular_momentum']

        if len(shell_am) > 1:
            raise RuntimeError("Cannot not remove diffuse functions from fused shell: {}".format(shell_am))

        shell_am = shell_am[0]
        if shell_am not in am_toremove:
            continue

        # Find the most diffuse function and remove it
        exponents = [(float(x), idx) for idx, x in enumerate(shell['exponents'])]
        sorted_exponents = sorted(exponents)
        diffuse_exponent, diffuse_idx = sorted_exponents[0]

        remove_primitive(shell, diffuse_idx)


def truhlar_calendarize(basis, month, use_copy=True):
    '''Create the truhlar "month" basis sets from the corresponding aug basis set

    In Papajak 2011, removal of diffuse function stopped before removal of s and p functions.
    This, conceivably, extends to d functions for transition metals.
    However, you can keep extending to further in the year by removing these functions,
    although it may affect stability. This is implemented in Gaussian with the t(month)
    basis sets - tjul, tjun, etc. The 'tmonth' and regular 'month' basis sets are equivalent
    until the 'maug' basis set is reached (containg no diffuse functions on H,He, s,p,d on transition
    metals, and s, p on other elements).

    Since the regular 'month' basis sets are equivalent until maug, we do not adopt the
    t(month) nomenclature. Instead, you can just go further backwards through the months
    until you run out of diffuse functions or run out of months.

    Parameters
    ----------
    basis : dict
        Basis set dictionary to work with
    month: str
        Month to create ('apr', 'jul', etc). Not case sensitive
    use_copy: bool
        If True, the input basis set is not modified.
    '''

    valid_months = ['jul', 'jun', 'may', 'apr', 'mar', 'feb', 'jan']
    month = month.lower()
    if month not in valid_months:
        raise RuntimeError("Month {} is not valid for truhlar calendarization")

    # We need to combine shells by AM
    # I guess we don't NEED to, but it makes things a lot easier
    basis = make_general(basis, use_copy=use_copy)

    # Find out he max am for the entire basis
    all_am = [misc.max_am(x['electron_shells']) for x in basis['elements'].values()]
    basis_max_am = max(all_am)

    # Figure out what to remove
    # We do this by finding the offset from jul
    # The index represents the offset, since index('jul') = 0
    # This offset represents how many diffuse functions to remove from each element
    month_offset = valid_months.index(month)

    # Are we removing all diffuse functions from all elements?
    # If so, we will end up with all elements corresponding to the cc-pVXZ basis.
    # While it is ok for some elements to end up that way, we don't want the entire
    # basis set to be like that
    # Note that for a given basis_max_am, there will be (basis_max_am+1) diffuse functions
    if month_offset > basis_max_am:  # or month_offset >= (basis_max_am+1)
        raise RuntimeError("Will not remove {} diffuse functions for basis with max am = {}".format(
            month_offset, basis_max_am))

    # jul = same as aug, except remove all diffuse from H,He
    # First, remove diffuse functions from H,He
    # This will always be true - we don't handle 'aug'
    for el_z in ['1', '2']:
        if el_z not in basis['elements']:
            continue

        eldata = basis['elements'][el_z]
        _element_remove_diffuse(eldata, 'all')

    # short circuit if we are doing jul
    if month_offset == 0:
        basis = prune_basis(basis, False)
        return basis

    # Now the rest
    for el_z, eldata in basis['elements'].items():
        if el_z == '1' or el_z == '2':
            continue

        _element_remove_diffuse(eldata, month_offset)

    basis = prune_basis(basis, False)
    return basis



'''
Testing of the BSE CLI interface
'''

import os
import subprocess
import pytest

from .common_testvars import fake_data_dir


def _test_cli_cmd(cmd):
    # NOTE: We do not enforce any encoding here. What is returned will be a byte string
    # For our purposes here, that is ok. We don't know what encoding is going to be
    # used (ie, windows)
    cmd = cmd.split(' ')
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT)


bse_cmds = [
    '-V', '-h', '--help', 'list-formats', 'list-reader-formats', 'list-ref-formats', 'list-roles', 'get-data-dir', 'list-basis-sets',
    'list-basis-sets -s sto', 'list-basis-sets -f duNNing', 'list-basis-sets -r orbiTAL -f ahlrichs',
    'list-basis-sets -e cn-OG', 'list-families', 'lookup-by-role def2-tzVp riFit', 'get-basis sto-3g nwchem',
    'get-basis cc-pvtz psi4 --elements=1-10', 'get-basis def2-tzvp turbomole --elements=H-9,11-Ar,cO',
    'get-basis cc-pvqz gaussian94 --version=1 --noheader', 'get-basis 6-31g nwchem --unc-gen --unc-spdf --unc-seg',
    'get-basis 6-31g nwchem --opt-gen', 'get-basis 6-31g nwchem --make-gen', 'get-refs 6-31g txt',
    'get-refs sto-3g bib --elements=6-Ne', 'get-refs ano-rcc txt --elements sc', 'get-info aug-cc-pv5z',
    'get-info def2-tzvp', 'get-notes sto-3g', 'get-family crenbl', 'get-versions cc-pvqz', 'get-family-notes sto'
]

fakebse_cmds = [
    'list-basis-sets', 'list-basis-sets -s fake', 'list-basis-sets -f bppfake', 'list-families',
    'lookup-by-role bppFAKEbasis jkFit', 'get-basis bppfakebasis nwchem', 'get-refs bppfakebasis txt',
    'get-info bppfakebasis', 'get-family bppfakebasis', 'get-versions bppfakebasis', 'get-family-notes bppfake'
]


@pytest.mark.parametrize('bse_cmd', bse_cmds)
def test_cli(bse_cmd):
    _test_cli_cmd('bse ' + bse_cmd)


@pytest.mark.parametrize('bse_cmd', fakebse_cmds)
def test_cli_datadir(bse_cmd):
    output = _test_cli_cmd('bse -d ' + fake_data_dir + ' ' + bse_cmd)
    assert b'bppfake' in output


def test_cli_createbundle_datadir(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5
    bfile_path = os.path.join(tmp_path, 'test_bundle_datadir.tar.bz2')
    output = _test_cli_cmd('bse -d ' + fake_data_dir + ' create-bundle gaussian94 bib ' + bfile_path)
    assert os.path.isfile(bfile_path)
    assert output.startswith(b'Created ')



'''
Testing of the BSE Curation CLI interface
'''

import os
import subprocess
import pytest
import shutil

from basis_set_exchange import fileio, curate
from .common_testvars import fake_data_dir, data_dir, curate_test_data_dir


def _test_curatecli_cmd(cmd):
    # NOTE: We do not enforce any encoding here. What is returned will be a byte string
    # For our purposes here, that is ok. We don't know what encoding is going to be
    # used (ie, windows)
    cmd = cmd.split(' ')
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT)

test_files1 = [ '6-31G.0.table.json', 'ahlrichs/def2-ECP.1.element.json', 'dunning/cc-pV5+dZ-add.1.json']
test_files1 = [os.path.join(data_dir, x) for x in test_files1]

test_files2 = [ 'dunning/cc-pV5+dZ-add.1.json', 'dunning/cc-pVDZ.1.json' ]
test_files2 = [os.path.join(data_dir, x) for x in test_files2]


bsecurate_cmds = [
    '-V', '-h', '--help',
    'elements-in-files ' + ' '.join(test_files1),
    'component-file-refs ' + ' '.join(test_files2)
]

fakebsecurate_cmds = [
    '-V', 'compare-basis-sets bppfakebasis bppfakebasis'
]


@pytest.mark.parametrize('bsecurate_cmd', bsecurate_cmds)
def test_curatecli(bsecurate_cmd):
    _test_curatecli_cmd('bsecurate ' + bsecurate_cmd)


@pytest.mark.parametrize('bsecurate_cmd', fakebsecurate_cmds)
def test_curatecli_datadir(bsecurate_cmd):
    _test_curatecli_cmd('bsecurate -d ' + fake_data_dir + ' ' + bsecurate_cmd)


def test_curatecli_makediff(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5

    filename1 = '6-31G_s_s-full.json.bz2'
    filename2 = '6-31G-full.json.bz2'

    file1 = os.path.join(curate_test_data_dir, filename1)
    file2 = os.path.join(curate_test_data_dir, filename2)

    tmpfile1 = os.path.join(tmp_path, filename1)
    tmpfile2 = os.path.join(tmp_path, filename2)

    shutil.copyfile(file1, tmpfile1)
    shutil.copyfile(file2, tmpfile2)

    _test_curatecli_cmd('bsecurate make-diff -l {} -r {}'.format(tmpfile1, tmpfile2))
    _test_curatecli_cmd('bsecurate make-diff -l {} -r {}'.format(tmpfile2, tmpfile1))

    diff1 = fileio.read_json_basis(tmpfile1 + '.diff')
    diff2 = fileio.read_json_basis(tmpfile2 + '.diff')

    assert len(diff1['elements']) == 36
    assert len(diff2['elements']) == 0

    reffilename = '6-31G_s_s-polarization.json.bz2'
    reffile = os.path.join(curate_test_data_dir, reffilename)
    refdata = fileio.read_json_basis(reffile)

    assert curate.compare_basis(diff1, refdata, rel_tol=0.0)


def test_curatecli_compare_1():
    output = _test_curatecli_cmd('bsecurate compare-basis-sets 6-31g 6-31g --version1 0 --version2 0')
    assert b"No difference found" in output

    output = _test_curatecli_cmd('bsecurate compare-basis-sets 6-31g 6-31g --version1 0 --version2 0 --uncontract-general')
    assert b"No difference found" in output

    output = _test_curatecli_cmd('bsecurate compare-basis-sets 6-31g 6-31g --version1 0 --version2 1')
    assert b"DIFFERENCES FOUND" in output


# yapf: disable
@pytest.mark.parametrize('filename1,filename2,expected',
                           [('6-31g-bse.gbs.bz2', '6-31g-bse.gbs.bz2', True),
                            ('6-31g-bse.gbs.bz2', '6-31g-bse.nw.bz2', True),
                            ('6-31g-bse.nw.bz2', '6-31g-bse-BAD1.gbs.bz2', False),
                            ('6-31g-bse.nw.bz2', '6-31g-bse-BAD2.gbs.bz2', False),
                            ('6-31g-bse.nw.bz2', '6-31g-bse-BAD3.gbs.bz2', False),
                            ('6-31g-bse.nw.bz2', '6-31g-bse-BAD4.gbs.bz2', False),
                            ('6-31g-bse.nw.bz2', '6-31g-bse-BAD5.gbs.bz2', False),
                            ('def2-ecp.gbs.bz2', 'def2-ecp.gbs.bz2', True),
                            ('def2-ecp.gbs.bz2', 'def2-ecp.nw.bz2', True),
                            ('def2-ecp.gbs.bz2', 'def2-ecp-BAD1.nw.bz2', False),
                            ('def2-ecp.gbs.bz2', 'def2-ecp-BAD2.nw.bz2', False),
                            ('def2-ecp.gbs.bz2', 'def2-ecp-BAD3.nw.bz2', False),
                            ('def2-ecp.gbs.bz2', 'def2-ecp-BAD4.nw.bz2', False),
                            ('def2-ecp.gbs.bz2', 'def2-ecp-BAD5.nw.bz2', False),
                            ('def2-ecp.gbs.bz2', 'def2-ecp-BAD6.gbs.bz2', False),
                            ('lanl2dz.nw.bz2', 'lanl2dz.nw.bz2', True),
                            ('lanl2dz.nw.bz2', 'lanl2dz-BAD1.nw.bz2', False),
                            ('lanl2dz.nw.bz2', 'lanl2dz-BAD2.nw.bz2', False)])
# yapf: enable
def test_curatecli_compare_files(filename1, filename2, expected):
    file1 = os.path.join(curate_test_data_dir, filename1)
    file2 = os.path.join(curate_test_data_dir, filename2)

    output = _test_curatecli_cmd('bsecurate compare-basis-files {} {} --uncontract-general'.format(file1, file2))
    if expected:
        assert b"No difference found" in output
    else:
        assert b"DIFFERENCES FOUND" in output

    output = _test_curatecli_cmd('bsecurate compare-basis-files {} {} --uncontract-general'.format(file2, file1))
    if expected:
        assert b"No difference found" in output
    else:
        assert b"DIFFERENCES FOUND" in output



'''
Conversion of basis sets to GAMESS-US
'''

from .. import lut, manip, sort, printing


def write_gamess_us_electron_basis(basis, electron_elements):
    # electronic part starts with $DATA
    s = '$DATA\n'

    for z in electron_elements:
        data = basis['elements'][z]

        el_name = lut.element_name_from_Z(z).upper()
        s += '\n' + el_name + "\n"

        for shell in data['electron_shells']:
            exponents = shell['exponents']
            coefficients = shell['coefficients']
            ncol = len(coefficients) + 2  #include index column
            nprim = len(exponents)

            am = shell['angular_momentum']
            amchar = lut.amint_to_char(am, hij=True, use_L=True).upper()
            s += '{}   {}\n'.format(amchar, nprim)

            # 1-based indexing
            idx_column = list(range(1, nprim + 1))
            point_places = [0] + [4 + 8 * i + 15 * (i - 1) for i in range(1, ncol)]
            s += printing.write_matrix([idx_column, exponents, *coefficients], point_places)

    s += "$END"
    return s


def write_gamess_us_ecp_basis(basis, ecp_elements, ecp_block=True):
    s = ""

    if ecp_block:
        s += "\n\n$ECP\n"

    for z in ecp_elements:
        data = basis['elements'][z]
        sym = lut.element_sym_from_Z(z).upper()
        max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])
        max_ecp_amchar = lut.amint_to_char([max_ecp_am], hij=True)

        # Sort lowest->highest, then put the highest at the beginning
        ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
        ecp_list.insert(0, ecp_list.pop())

        s += '{}-ECP GEN    {}    {}\n'.format(sym, data['ecp_electrons'], max_ecp_am)

        for pot in ecp_list:
            rexponents = pot['r_exponents']
            gexponents = pot['gaussian_exponents']
            coefficients = pot['coefficients']
            nprim = len(rexponents)

            am = pot['angular_momentum']
            amchar = lut.amint_to_char(am, hij=False)

            # Title line
            if am[0] == max_ecp_am:
                s += '{:<5} ----- {}-ul potential -----\n'.format(nprim, amchar)
            else:
                s += '{:<5} ----- {}-{} potential -----\n'.format(nprim, amchar, max_ecp_amchar)

            point_places = [8, 23, 32]
            s += printing.write_matrix([*coefficients, rexponents, gexponents], point_places)

    if ecp_block:
        s += "$END\n"
    return s


def write_gamess_us_common(basis, ecp_func):
    '''Converts the electronic basis to GAMESS-US, using a
       different function for ECP
    '''

    s = ''

    # Uncontract all but SP
    basis = manip.uncontract_general(basis, True)
    basis = manip.uncontract_spdf(basis, 1, False)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        s += write_gamess_us_electron_basis(basis, electron_elements)

    # Write out ECP
    if ecp_elements:
        s += ecp_func(basis, ecp_elements)

    return s


def write_gamess_us(basis):
    '''Converts a basis set to GAMESS-US
    '''

    return write_gamess_us_common(basis, write_gamess_us_ecp_basis)



'''
Some data common to all tests
'''

import os
import basis_set_exchange as bse

# The directory containing this file
_my_dir = os.path.dirname(os.path.abspath(__file__))

# Use random for getting sets of elements
rand_seed = 39466  # from random.org

# Load all the metadata once
data_dir = bse.api._default_data_dir
bs_metadata = bse.get_metadata()
bs_names = list(bs_metadata.keys())
bs_read_formats = list(bse.get_reader_formats().keys())
bs_write_formats = list(bse.get_writer_formats().keys()) + [None]
ref_formats = list(bse.get_reference_formats().keys()) + [None]
all_families = bse.get_families()
all_roles = bse.get_roles()
true_false = [True, False]

# All basis names/versions combinations
bs_names_vers = []
for k, v in bs_metadata.items():
    for ver in v['versions'].keys():
        bs_names_vers.append((k, ver))

# Directory with some fake data
fake_data_dir = os.path.join(_my_dir, 'fakedata')

# Directory with authoritative sources
auth_data_dir = os.path.join(_my_dir, 'sources')

# Directory with files for testing curation functions
curate_test_data_dir = os.path.join(_my_dir, 'curate_test_data')

# Directory with files for testing readers
reader_test_data_dir = os.path.join(_my_dir, 'reader_test_data')

# Directory with files for testing the validator
validator_test_data_dir = os.path.join(_my_dir, 'validator_test_data')

# Directory with files for testing dunning aug extension
dunningext_test_data_dir = os.path.join(_my_dir, 'dunning_extend')

# Directory with files for testing truhlar calenderization
truhlar_test_data_dir = os.path.join(_my_dir, 'truhlar')

# All files in the data dir
all_files = bse.fileio.get_all_filelist(data_dir)
all_metadata_files = all_files[0]
all_table_files = all_files[1]
all_element_files = all_files[2]
all_component_files = all_files[3]

# Full paths to all the files
all_file_paths = [[os.path.join(data_dir, x) for x in y] for y in all_files]
all_metadata_paths = all_file_paths[0]
all_table_paths = all_file_paths[1]
all_element_paths = all_file_paths[2]
all_component_paths = all_file_paths[3]

# A representative sample of basis sets
bs_names_sample = ['6-31g', '6-31+g*', 'aug-cc-pvtz', 'lanl2dz', 'def2-tzvp', 'jorge-tzp', 'sto-3g', 'fano-qz']


# Authoritative data in the sources dir
auth_src_map = {}
for x in os.listdir(auth_data_dir):
    if not x.endswith('.bz2'):
        continue

    # remove .fmt.bz2
    base, _ = os.path.splitext(x)
    base, _ = os.path.splitext(base)

    if base in auth_src_map:
        raise RuntimeError("Duplicate basis set in authoritative sources: {}".format(base))

    auth_src_map[base] = os.path.join(auth_data_dir, x)


def bool_matrix(size):
    '''Returns an identity matrix of a given size consisting of bool types
    '''
    ret = [[False for i in range(size)] for j in range(size)]
    for x in range(size):
        ret[x][x] = True
    return ret



'''
Helper functions for writing out basis set in various formats
'''


def find_range(coeffs):
    '''
    Find the range in a list of coefficients where the coefficient is nonzero
    '''

    coeffs = [float(x) != 0 for x in coeffs]
    first = coeffs.index(True)
    coeffs.reverse()
    last = len(coeffs) - coeffs.index(True) - 1
    return first, last
